@@ -3,104 +3,206 @@
  */
 
 #include "sh.h"
+#include "ksh_stat.h"
 #include <ctype.h>
 
 int
 c_cd(wp)
 	char	**wp;
 {
-	int		rval;
-	int		cdnode;		/* was a node from cdpath added in? */
-	int		printpath = 0;	/* print where we cd'd? */
-	char		*cp, *cdpath, *pwd, *oldpwd;
-	struct tbl	*pwd_s, *oldpwd_s;
-	char		path[PATH];
-	extern int	make_path();
-	extern void	simplify_path();
+	int optc;
+	int physical = Flag(FPHYSICAL);
+	int cdnode;			/* was a node from cdpath added in? */
+	int printpath = 0;		/* print where we cd'd? */
+	int rval;
+	struct tbl *pwd_s, *oldpwd_s;
+	XString xs;
+	char *xp;
+	char *dir, *try, *pwd;
+	int phys_path;
+	char *cdpath;
+
+	while ((optc = ksh_getopt(wp, &builtin_opt, "LP")) != EOF)
+		switch (optc) {
+		case 'L':
+			physical = 0;
+			break;
+		case 'P':
+			physical = 1;
+			break;
+		case '?':
+			return 1;
+		}
+	wp += builtin_opt.optind;
 
 	if (Flag(FRESTRICTED)) {
 		bi_errorf("restricted shell - can't cd");
 		return 1;
 	}
-	pwd = strval(pwd_s = global("PWD"));
-	oldpwd = strval(oldpwd_s = global("OLDPWD"));
 
-	if ((cp = wp[1]) == (char *) 0) {
-		if ((cp = strval(global("HOME"))) == null) {
-			bi_errorf("no home directory");
+	pwd_s = global("PWD");
+	oldpwd_s = global("OLDPWD");
+
+	if (!wp[0]) {
+		/* No arguments - go home */
+		if ((dir = strval(global("HOME"))) == null) {
+			bi_errorf("no home directory (HOME not set)");
 			return 1;
 		}
-	} else if (wp[2] == (char *) 0) {
-		if (strcmp(cp, "-") == 0) {
-			if ((cp = oldpwd) == null) {
+	} else if (!wp[1]) {
+		/* One argument: - or dir */
+		dir = wp[0];
+		if (strcmp(dir, "-") == 0) {
+			dir = strval(oldpwd_s);
+			if (dir == null) {
 				bi_errorf("no OLDPWD");
 				return 1;
 			}
 			printpath++;
 		}
-	} else if (wp[3] == (char *) 0) {
-		int	ilen, olen, nlen;
+	} else if (!wp[2]) {
+		/* Two arguments - substitute arg1 in PWD for arg2 */
+		int ilen, olen, nlen, elen;
+		char *cp;
 
+		if (!current_wd[0]) {
+			bi_errorf("don't know current directory");
+			return 1;
+		}
 		/* substitue arg1 for arg2 in current path.
 		 * if the first substitution fails because the cd fails
 		 * we could try to find another substitution. For now
 		 * we don't
 		 */
-		if ((cp = strstr(pwd, wp[1])) == (char *) 0) {
+		if ((cp = strstr(current_wd, wp[0])) == (char *) 0) {
 			bi_errorf("bad substitution");
 			return 1;
 		}
-		ilen = cp - pwd;
-		olen = strlen(wp[1]);
-		nlen = strlen(wp[2]);
-		if (ilen + nlen + strlen(pwd + ilen + olen) + 1 > sizeof(path))
-		{
-			bi_errorf("path too long");
-			return 1;
-		}
-		strncpy(path, pwd, ilen);
-		strcpy(path + ilen, wp[2]);
-		strcpy(path + ilen + nlen, pwd + ilen + olen);
-		cp = path;
+		ilen = cp - current_wd;
+		olen = strlen(wp[0]);
+		nlen = strlen(wp[1]);
+		elen = strlen(current_wd + ilen + olen) + 1;
+		dir = alloc(ilen + nlen + elen, ATEMP);
+		memcpy(dir, current_wd, ilen);
+		memcpy(dir + ilen, wp[1], nlen);
+		memcpy(dir + ilen + nlen, current_wd + ilen + olen, elen);
 		printpath++;
-		/* XXX: make_path() copies cp to path, in this case cp to cp. */
 	} else {
 		bi_errorf("too many arguments");
 		return 1;
 	}
 
+	Xinit(xs, xp, PATH, ATEMP);
+	/* xp will have a bogus value after make_path() - set it to 0
+	 * so that if it's used, it will cause a dump
+	 */
+	xp = (char *) 0;
+
 	cdpath = strval(global("CDPATH"));
 	do {
-		cdnode = make_path(pwd, cp, &cdpath, path, sizeof(path));
-		simplify_path(path);
-		rval = chdir(path);
+		cdnode = make_path(current_wd, dir, &cdpath, &xs, &phys_path);
+#ifdef S_ISLNK
+		if (physical)
+			rval = chdir(try = Xstring(xs, xp) + phys_path);
+		else
+#endif /* S_ISLNK */
+		{
+			simplify_path(Xstring(xs, xp));
+			rval = chdir(try = Xstring(xs, xp));
+		}
 	} while (rval < 0 && cdpath != (char *) 0);
 
 	if (rval < 0) {
-		bi_errorf("%s: bad directory", cp);
+		if (cdnode)
+			bi_errorf("%s: bad directory", dir);
+		else
+			bi_errorf("%s - %s", try, strerror(errno));
 		return 1;
 	}
 
+	/* Clear out tracked aliases with relative paths */
 	flushcom(0);
 
-	if (printpath || cdnode)
-		shprintf("%s\n", path);
+	/* Set OLDPWD */
+	if (current_wd[0])
+		setstr(oldpwd_s, current_wd);
 
+	if (!ISABSPATH(Xstring(xs, xp))) {
 #ifdef OS2
-	/* OS2 is like MSDOS in that paths can be relative to a drive
-	 * specification, hence, the above code doesn't always calculate
-	 * the current directory correctly.  Workaround is to let getcwd()
-	 * figure it out.
-	 */
-	getcwd(path, sizeof(path));
+		/* simplify_path() doesn't know about os/2's drive contexts,
+		 * so it can't set current_wd when changing to a:foo.
+		 * Handle this by calling getcwd()...
+		 */
+		pwd = ksh_get_wd((char *) 0, 0);
+#else /* OS2 */
+		pwd = (char *) 0;
 #endif /* OS2 */
+	} else
+#ifdef S_ISLNK
+	if (!physical || !(pwd = get_phys_path(Xstring(xs, xp))))
+#endif /* S_ISLNK */
+		pwd = Xstring(xs, xp);
 
-	setstr(oldpwd_s, pwd);
-	setstr(pwd_s, path);
+	/* Set PWD */
+	if (pwd) {
+		set_current_wd(pwd);
+		setstr(pwd_s, pwd);
+	} else {
+		set_current_wd(null);
+		pwd = Xstring(xs, xp);
+		/* XXX unset $PWD? */
+	}
+	if (printpath || cdnode)
+		shprintf("%s\n", pwd);
 
 	return 0;
 }
 
+int
+c_pwd(wp)
+	char	**wp;
+{
+	int optc;
+	int physical = Flag(FPHYSICAL);
+	char *p;
+
+	while ((optc = ksh_getopt(wp, &builtin_opt, "LP")) != EOF)
+		switch (optc) {
+		case 'L':
+			physical = 0;
+			break;
+		case 'P':
+			physical = 1;
+			break;
+		case '?':
+			return 1;
+		}
+	wp += builtin_opt.optind;
+
+	if (wp[0]) {
+		bi_errorf("too many arguments");
+		return 1;
+	}
+#ifdef S_ISLNK
+	p = current_wd[0] ? (physical ? get_phys_path(current_wd) : current_wd)
+			  : (char *) 0;
+#else /* S_ISLNK */
+	p = current_wd[0] ? current_wd : (char *) 0;
+#endif /* S_ISLNK */
+	if (p && eaccess(p, R_OK) < 0)
+		p = (char *) 0;
+	if (!p) {
+		p = ksh_get_wd((char *) 0, 0);
+		if (!p) {
+			bi_errorf("can't get current directory - %s",
+				strerror(errno));
+			return 1;
+		}
+	}
+	shprintf("%s\n", p);
+	return 0;
+}
+
 int
 c_print(wp)
 	register char **wp;
@@ -333,7 +435,7 @@ c_whence(wp)
 
 	while ((vflag || ret == 0) && (id = *wp++) != NULL) {
 		tp = NULL;
-		if (iam_whence && !pflag)
+		if ((iam_whence || vflag) && !pflag)
 			tp = tsearch(&keywords, id, hash(id));
 		if (!tp && !pflag) {
 			tp = tsearch(&aliases, id, hash(id));
@@ -348,7 +450,7 @@ c_whence(wp)
 		switch (tp->type) {
 		  case CKEYWD:
 			if (vflag)
-				shprintf(" is a keyword");
+				shprintf(" is a reserved word");
 			break;
 		  case CALIAS:
 			if (vflag)
@@ -363,7 +465,7 @@ c_whence(wp)
 			if (vflag) {
 				shprintf(" is a");
 				if (tp->flag & EXPORT)
-					shprintf(" exported");
+					shprintf("n exported");
 				if (tp->flag & TRACE)
 					shprintf(" traced");
 				if (!(tp->flag & ISSET)) {
@@ -849,12 +951,17 @@ c_let(wp)
 	char **wp;
 {
 	int rv = 1;
+	long val;
 
 	if (wp[1] == (char *) 0) /* at&t ksh does this */
 		bi_errorf("no arguments");
 	else
 		for (wp++; *wp; wp++)
-			rv = evaluate(*wp) == 0;
+			if (!evaluate(*wp, &val, TRUE)) {
+				rv = 2;	/* distinguish error from zero result */
+				break;
+			} else
+				rv = val == 0;
 	return rv;
 }
 
@@ -1069,7 +1176,8 @@ getopts_reset(val)
 	int val;
 {
 	if (!getopts_noset && val >= 1) {
-		ksh_getopt_reset(&user_opt, Flag(FPOSIX) ? 0 : GF_PLUSOPT);
+		ksh_getopt_reset(&user_opt,
+			GI_NONAME | (Flag(FPOSIX) ? 0 : GF_PLUSOPT));
 		user_opt.optind = val;
 	}
 }
@@ -1143,6 +1251,8 @@ c_getopts(wp)
 	vq = global(var);
 	if (vq->flag & RDONLY)
 		bi_errorf("%s is readonly", var);
+	if (Flag(FEXPORT))
+		typeset(var, EXPORT, 0, 0, 0);
 	setstr(vq, buf);
 
 	getopts_noset = 1;
@@ -1150,7 +1260,7 @@ c_getopts(wp)
 	getopts_noset = 0;
 
 	if (user_opt.optarg == (char *) 0)
-		unset(global("OPTARG"));
+		unset(global("OPTARG"), 0);
 	else
 		setstr(global("OPTARG"), user_opt.optarg);
 
@@ -1165,12 +1275,15 @@ int
 c_bind(wp)
 	register char **wp;
 {
-	int rv = 0, macro = 0;
+	int rv = 0, macro = 0, list = 0;
 	register char *cp;
 	int optc;
 
-	while ((optc = ksh_getopt(wp, &builtin_opt, "m")) != EOF)
+	while ((optc = ksh_getopt(wp, &builtin_opt, "lm")) != EOF)
 		switch (optc) {
+		  case 'l':
+			list = 1;
+			break;
 		  case 'm':
 			macro = 1;
 			break;
@@ -1180,13 +1293,13 @@ c_bind(wp)
 	wp += builtin_opt.optind;
 
 	if (*wp == NULL)	/* list all */
-		rv = x_bind((char*)NULL, (char*)NULL, 0);
+		rv = x_bind((char*)NULL, (char*)NULL, 0, list);
 
 	for (; *wp != NULL; wp++) {
 		cp = strchr(*wp, '=');
 		if (cp != NULL)
 			*cp++ = '\0';
-		if (x_bind(*wp, cp, macro))
+		if (x_bind(*wp, cp, macro, 0))
 			rv = 1;
 	}
 
@@ -1213,6 +1326,7 @@ const struct builtin kshbuiltins [] = {
 	{"+kill", c_kill},
 	{"let", c_let},
 	{"print", c_print},
+	{"pwd", c_pwd},
  	{"*=readonly", c_typeset},
 	{"=typeset", c_typeset},
 	{"+unalias", c_unalias},
