@@ -11,10 +11,10 @@
 #define EXTERN
 #include "edit.h"
 #undef EXTERN
-#ifdef M_UNIX	/* SCO Unix 3.2v4.1 */
+#ifdef OS_SCO	/* SCO Unix 3.2v4.1 */
 # include <sys/stream.h>	/* needed for <sys/ptem.h> */
 # include <sys/ptem.h>		/* needed for struct winsize */
-#endif /* M_UNIX */
+#endif /* OS_SCO */
 
 static char	vdisable_c;
 
@@ -33,7 +33,7 @@ x_init()
 		struct winsize ws;
 
 		if (ioctl(tty_fd, TIOCGWINSZ, &ws) >= 0 && ws.ws_col) {
-			x_cols = ws.ws_col;
+			x_cols = ws.ws_col < MIN_COLS ? MIN_COLS : ws.ws_col;
 			setint(global("COLUMNS"), (long) x_cols);
 		}
 	}
@@ -98,7 +98,7 @@ x_getc()
 	char c;
 	int n;
 
-	while ((n = read(0, &c, 1)) < 0 && errno == EINTR)
+	while ((n = blocking_read(0, &c, 1)) < 0 && errno == EINTR)
 		if (trap) {
 			x_mode(FALSE);
 			runtraps(0);
@@ -142,7 +142,7 @@ x_mode(onoff)
 		return x_cur_mode;
 	prev = x_cur_mode;
 	x_cur_mode = onoff;
-#ifndef OS2
+
 	if (onoff) {
 		TTY_state	cb;
 		X_chars		oldchars;
@@ -150,46 +150,46 @@ x_mode(onoff)
 		oldchars = edchars;
 		cb = tty_state;
 
-# if defined(HAVE_TERMIOS_H) || defined(HAVE_TERMIO_H)
+#if defined(HAVE_TERMIOS_H) || defined(HAVE_TERMIO_H)
 		edchars.erase = cb.c_cc[VERASE];
 		edchars.kill = cb.c_cc[VKILL];
 		edchars.intr = cb.c_cc[VINTR];
 		edchars.quit = cb.c_cc[VQUIT];
 		edchars.eof = cb.c_cc[VEOF];
-#  ifdef VWERASE
+# ifdef VWERASE
 		edchars.werase = cb.c_cc[VWERASE];
-#  endif
-#  ifdef _CRAY2		/* brain-damaged terminal handler */
+# endif
+# ifdef _CRAY2		/* brain-damaged terminal handler */
 		cb.c_lflag &= ~(ICANON|ECHO);
 		/* rely on print routine to map '\n' to CR,LF */
-#  else
+# else
 		cb.c_iflag &= ~(INLCR|ICRNL);
-#   ifdef _BSD_SYSV	/* need to force CBREAK instead of RAW (need CRMOD on output) */
+#  ifdef _BSD_SYSV	/* need to force CBREAK instead of RAW (need CRMOD on output) */
 		cb.c_lflag &= ~(ICANON|ECHO);
-#   else
-#    ifdef SWTCH	/* need CBREAK to handle swtch char */
+#  else
+#   ifdef SWTCH	/* need CBREAK to handle swtch char */
 		cb.c_lflag &= ~(ICANON|ECHO);
 		cb.c_lflag |= ISIG;
 		cb.c_cc[VINTR] = vdisable_c;
 		cb.c_cc[VQUIT] = vdisable_c;
-#    else
+#   else
 		cb.c_lflag &= ~(ISIG|ICANON|ECHO);
-#    endif
 #   endif
-#   ifdef VLNEXT
+#  endif
+#  ifdef VLNEXT
 		/* osf/1 processes lnext when ~icanon */
 		cb.c_cc[VLNEXT] = vdisable_c;
-#   endif /* VLNEXT */
+#  endif /* VLNEXT */
 		cb.c_cc[VTIME] = 0;
 		cb.c_cc[VMIN] = 1;
-#  endif	/* _CRAY2 */
-# else
+# endif	/* _CRAY2 */
+#else
 	/* Assume BSD tty stuff. */
 		edchars.erase = cb.sgttyb.sg_erase;
 		edchars.kill = cb.sgttyb.sg_kill;
 		cb.sgttyb.sg_flags &= ~ECHO;
 		cb.sgttyb.sg_flags |= CBREAK;
-#   ifdef TIOCGATC
+#  ifdef TIOCGATC
 		edchars.intr = cb.lchars.tc_intrc;
 		edchars.quit = cb.lchars.tc_quitc;
 		edchars.eof = cb.lchars.tc_eofc;
@@ -201,33 +201,33 @@ x_mode(onoff)
 		cb.lchars.tc_intrc = -1;
 		cb.lchars.tc_quitc = -1;
 		cb.lchars.tc_rprntc = -1;
-#   else
+#  else
 		edchars.intr = cb.tchars.t_intrc;
 		edchars.quit = cb.tchars.t_quitc;
 		edchars.eof = cb.tchars.t_eofc;
 		cb.tchars.t_intrc = -1;
 		cb.tchars.t_quitc = -1;
-#    ifdef TIOCGLTC
+#   ifdef TIOCGLTC
 		edchars.werase = cb.ltchars.t_werasc;
 		cb.ltchars.t_suspc = -1;
 		cb.ltchars.t_dsuspc = -1;
 		cb.ltchars.t_lnextc = -1;
 		cb.ltchars.t_rprntc = -1;
-#    endif
-#   endif /* TIOCGATC */
-# endif /* HAVE_TERMIOS_H || HAVE_TERMIO_H */
+#   endif
+#  endif /* TIOCGATC */
+#endif /* HAVE_TERMIOS_H || HAVE_TERMIO_H */
 
 		set_tty(tty_fd, &cb, TF_WAIT);
 
 		if (memcmp(&edchars, &oldchars, sizeof(edchars)) != 0) {
-# ifdef EMACS
+#ifdef EMACS
 			x_emacs_keys(&edchars);
-# endif
+#endif
 		}
 	} else
 		/* TF_WAIT doesn't seem to be necessary when leaving xmode */
 		set_tty(tty_fd, &tty_state, TF_NONE);
-#endif /* !OS2 */
+
 	return prev;
 }
 
@@ -245,47 +245,41 @@ x_mode(onoff)
  
 int
 promptlen(cp)
-  register char  *cp;
+    register char  *cp;
 {
-  register int count = 0;
+    register int count = 0;
 
-  while (*cp)
-  {
-    if ( *cp == '\n' || *cp == '\r' )
-      {
-	count = 0;
-	cp++;
-      }
-    else if ( *cp == '\t' )
-      {
-	count = (count | 7) + 1;
-	cp++;
-      }
-    else if ( *cp == '\b' )
-      {
-	if (count > 0)
+    while (*cp) {
+	if (*cp == '\n' || *cp == '\r') {
+	    count = 0;
+	    cp++;
+	} else if (*cp == '\t') {
+	    count = (count | 7) + 1;
+	    cp++;
+	} else if (*cp == '\b') {
+	    if (count > 0)
 		count--;
-	cp++;
-      }
-    else if ( *cp++ != '!' )
-      count++;
-    else if ( *cp == '!' )
-      {
-	cp++;
-	count++;
-      }
-      else
-      {
-	register int i = source->line + 1;
-
-	do
-	{
-	  count ++;
+	    cp++;
 	}
-	while( ( i /= 10 ) > 0 );
-      }
-  }
-  return count;
+#if 1
+	else
+	  cp++, count++;
+#else
+	else if (*cp++ != '!')
+	  count++;
+	else if (*cp == '!') {
+	    cp++;
+	    count++;
+	} else {
+	    register int i = source->line + 1;
+
+	    do
+		count++;
+	    while ((i /= 10) > 0);
+	}
+#endif /* 1 */
+    }
+    return count;
 }
 
 void
