@@ -180,6 +180,11 @@ expand(cp, wp, f)
 
 	if (cp == NULL)
 		internal_errorf(1, "expand(NULL)");
+	/* for alias, readonly, set, typeset commands */
+	if ((f & DOVACHECK) && is_wdvarassign(cp)) {
+		f &= ~(DOVACHECK|DOBLANK|DOGLOB|DOTILDE);
+		f |= DOASNTILDE;
+	}
 	if (Flag(FNOGLOB))
 		f &= ~DOGLOB;
 #ifdef BRACEEXPAND
@@ -256,9 +261,10 @@ expand(cp, wp, f)
 					char *p;
 
 					v.flag = DEFINED|ISSET|INTEGER;
-					v.type = 0;
+					v.type = 10; /* not default */
 					v.name[0] = '\0';
-					v.val.i = evaluate(substitute(sp, 0));
+					v_evaluate(&v, substitute(sp, 0),
+						FALSE);
 					sp = strchr(sp, 0) + 1;
 					for (p = strval(&v); *p; ) {
 						Xcheck(ds, dp);
@@ -688,7 +694,7 @@ varsub(xp, sp, word, stypep)
 		/* Check for size of array */
 		if ((p=strchr(sp,'[')) && (p[1]=='*'||p[1]=='@') && p[2]==']') {
 			c = 0;
-			vp = global(basename(sp));
+			vp = global(arrayname(sp));
 			if (vp->flag & (ISSET|ARRAY))
 				zero_ok = 1;
 			for (; vp; vp = vp->u.array)
@@ -756,7 +762,7 @@ varsub(xp, sp, word, stypep)
 				return -1;
 			}
 			XPinit(wv, 32);
-			vp = global(basename(sp));
+			vp = global(arrayname(sp));
 			for (; vp; vp = vp->u.array) {
 				if (!(vp->flag&ISSET))
 					continue;
@@ -790,9 +796,8 @@ varsub(xp, sp, word, stypep)
 	    (((stype&0x80) ? *xp->str=='\0' : xp->str==null) ? /* undef? */
 	     c == '=' || c == '-' || c == '?' : c == '+'))
 		state = XBASE;	/* expand word instead of variable value */
-	if (Flag(FNOUNSET) && (ctype(c, C_SUBOP2)
-			       || (state != XBASE && xp->str == null
-				   && c != '+')))
+	if (Flag(FNOUNSET) && xp->str == null
+	    && (ctype(c, C_SUBOP2) || (state != XBASE && c != '+')))
 		errorf("%s: unset variable", sp);
 	return state;
 }
@@ -865,7 +870,7 @@ trimsub(str, pat, how)
 	  case '#':		/* shortest at begin */
 		for (p = str; p <= end; p++) {
 			c = *p; *p = '\0';
-			if (gmatch(str, pat)) {
+			if (gmatch(str, pat, FALSE)) {
 				*p = c;
 				return p;
 			}
@@ -875,7 +880,7 @@ trimsub(str, pat, how)
 	  case '#'|0x80:	/* longest match at begin */
 		for (p = end; p >= str; p--) {
 			c = *p; *p = '\0';
-			if (gmatch(str, pat)) {
+			if (gmatch(str, pat, FALSE)) {
 				*p = c;
 				return p;
 			}
@@ -884,7 +889,7 @@ trimsub(str, pat, how)
 		break;
 	  case '%':		/* shortest match at end */
 		for (p = end; p >= str; p--) {
-			if (gmatch(p, pat)) {
+			if (gmatch(p, pat, FALSE)) {
 				c = *p; *p = '\0';
 				match = strsave(str, ATEMP);
 				*p = c;
@@ -894,7 +899,7 @@ trimsub(str, pat, how)
 		break;
 	  case '%'|0x80:	/* longest match at end */
 		for (p = str; p <= end; p++) {
-			if (gmatch(p, pat)) {
+			if (gmatch(p, pat, FALSE)) {
 				c = *p; *p = '\0';
 				match = strsave(str, ATEMP);
 				*p = c;
@@ -943,6 +948,7 @@ globit(xs, xpp, sp, wp, check)
 {
 	register char *np;	/* next source component */
 	char *xp = *xpp;
+	char odirsep;
 
 	/* This to allow long expansions to be interrupted */
 	intrcheck();
@@ -1005,8 +1011,11 @@ globit(xs, xpp, sp, wp, check)
 		*xp++ = *sp++;
 	}
 	np = strchr_dirsep(sp);
-	if (np != NULL)
+	if (np != NULL) {
+		odirsep = *np;	/* don't assume DIRSEP, can be multiple kinds */
 		*np++ = '\0';
+	} else
+		odirsep = '\0'; /* keep gcc quiet */
 
 	*xpp = xp;
 
@@ -1036,7 +1045,8 @@ globit(xs, xpp, sp, wp, check)
 			if (name[0] == '.' &&
 			    (name[1] == 0 || (name[1] == '.' && name[2] == 0)))
 				continue; /* always ignore . and .. */
-			if ((*name == '.' && *sp != '.') || !gmatch(name, sp))
+			if ((*name == '.' && *sp != '.')
+			    || !gmatch(name, sp, TRUE))
 				continue;
 
 			len = NLENGTH(d) + 1;
@@ -1051,7 +1061,7 @@ globit(xs, xpp, sp, wp, check)
 	}
 
 	if (np != NULL)
-		*--np = DIRSEP;
+		*--np = odirsep;
 }
 
 /* Check if p contains something that needs globbing; if it does, 0 is
@@ -1135,7 +1145,7 @@ maybe_expand_tilde(p, dsp, dpp, isassign)
 		p += 2;
 	}
 	*tp = '\0';
-	r = (p[0] == EOS || p[0] == CHAR) ? tilde(Xstring(ts, tp)) : (char *) 0;
+	r = (p[0] == EOS || p[0] == CHAR || p[0] == CSUBST) ? tilde(Xstring(ts, tp)) : (char *) 0;
 	Xfree(ts, tp);
 	if (r) {
 		while (*r) {
