@@ -139,25 +139,30 @@ static struct tbl *intvar   ARGS((struct tbl *vp));
 /*
  * parse and evalute expression
  */
-long
-evaluate(expr)
+int
+evaluate(expr, rval, error_ok)
 	const char *expr;
+	long *rval;
+	int error_ok;
 {
 	struct tbl v;
+	int ret;
 
 	v.flag = DEFINED|INTEGER;
 	v.type = 0;
-	v_evaluate(&v, expr);
-	return v.val.i;
+	ret = v_evaluate(&v, expr, error_ok);
+	*rval = v.val.i;
+	return ret;
 }
 
 /*
  * parse and evalute expression, storing result in vp.
  */
-void
-v_evaluate(vp, expr)
+int
+v_evaluate(vp, expr, error_ok)
 	struct tbl *vp;
 	const char *expr;
+	volatile int error_ok;
 {
 	struct tbl *v;
 	Expr_state curstate;
@@ -173,6 +178,11 @@ v_evaluate(vp, expr)
 	if ((i = setjmp(e->jbuf))) {
 		quitenv();
 		es = curstate.prev;
+		if (i == LAEXPR) {
+			if (error_ok)
+				return 0;
+			errorf((char *) 0);
+		}
 		unwind(i);
 		/*NOTREACHED*/
 	}
@@ -196,6 +206,8 @@ v_evaluate(vp, expr)
 
 	es = curstate.prev;
 	quitenv();
+
+	return 1;
 }
 
 static void
@@ -226,24 +238,24 @@ evalerr(type, str)
 		default:
 			s = opinfo[(int)es->tok].name;
 		}
-		/* XXX have errorf() just go back to error handler? */
-		errorf("%s: unexpected `%s'", es->expression, s);
+		warningf(TRUE, "%s: unexpected `%s'", es->expression, s);
 		break;
 
 	case ET_BADLIT:
-		errorf("%s: bad number `%s'", es->expression, str);
+		warningf(TRUE, "%s: bad number `%s'", es->expression, str);
 		break;
 
 	case ET_BADVAR:
-		errorf("%s: value of variable `%s' not a number",
+		warningf(TRUE, "%s: value of variable `%s' not a number",
 			es->expression, str);
 		break;
 
 	default: /* keep gcc happy */
 	case ET_STR:
-		errorf("%s: %s", es->expression, str);
+		warningf(TRUE, "%s: %s", es->expression, str);
 		break;
 	}
+	unwind(LAEXPR);
 }
 
 static struct tbl *
