@@ -57,38 +57,42 @@ static int sprinkle ARGS((int));
 
 # endif	/* of EASY_HISTORY */
 
-static char   **histget ARGS((char *str, int approx));
+static int	hist_execute ARGS((char *cmd));
+static int	hist_replace ARGS((char **hp, char *pat, char *rep,
+				   int global));
+static char   **hist_get ARGS((char *str, int approx));
+static char   **hist_get_newest ARGS((int print_err));
+static char   **hist_get_oldest ARGS((int print_err));
 static void	histbackup ARGS((void));
-static char    *histrpl ARGS((char *s, char *pat, char *rep, int global));
 
 static char   **current;	/* current postition in history[] */
 static int	curpos;		/* current index in history[] */
 static char    *hname;		/* current name of history file */
 static int	hstarted;	/* set after hist_init() called */
+static Source	*hist_source;	
 
 
 int
 c_fc(wp)
 	register char **wp;
 {
-	register char *id;
 	struct shf *shf;
 	struct temp UNINITIALIZED(*tf);
-	register char **hp;
-	char **hbeg, **hend;
-	char *p, *cmd = NULL;
-	int lflag = 0, nflag = 0, sflag = 0, rflag = 0, gflag = 0;
+	char *p, *editor = (char *) 0;
+	int gflag = 0, lflag = 0, nflag = 0, sflag = 0, rflag = 0;
 	int optc;
+	char *first = (char *) 0, *last = (char *) 0;
+	char **hfirst, **hlast, **hp;
 
-	while ((optc = ksh_getopt(wp, &builtin_opt, "e:glnr")) != EOF)
+	while ((optc = ksh_getopt(wp, &builtin_opt, "e:glnrs0,1,2,3,4,5,6,7,8,9,")) != EOF)
 		switch (optc) {
 		  case 'e':
 			p = builtin_opt.optarg;
 			if (strcmp(p, "-") == 0)
 				sflag++;
 			else {
-				cmd = strnsave(p, strlen(p) + 4, ATEMP);
-				strcat(cmd, " $_");
+				editor = strnsave(p, strlen(p) + 4, ATEMP);
+				strcat(editor, " $_");
 			}
 			break;
 		  case 'g': /* non-at&t ksh */
@@ -103,161 +107,330 @@ c_fc(wp)
 		  case 'r':
 			rflag++;
 			break;
+		  case 's':	/* posix version of -e - */
+			sflag++;
+			break;
+		  /* kludge city - accept -num as -- -num (kind of) */
+		  case '0': case '1': case '2': case '3': case '4':
+		  case '5': case '6': case '7': case '8': case '9':
+			p = shf_smprintf("-%c%s",
+					optc, builtin_opt.optarg);
+			if (!first)
+				first = p;
+			else if (!last)
+				last = p;
+			else {
+				bi_errorf("too many arguments");
+				return 1;
+			}
+			break;
 		  case '?':
 			return 1;
 		}
 	wp += builtin_opt.optind;
 
+	/* Substitute and execute command */
 	if (sflag) {
-		char *pat = NULL, *rep = NULL;
+		char *pat = (char *) 0, *rep = (char *) 0;
 
-		hp = histptr - 1;
-		while ((id = *wp++) != NULL) {
-			/* todo: multiple substitutions */
-			if (*id && (p = strchr(id + 1, '=')) != NULL) {
-				pat = id;
+		if (editor || lflag || nflag || rflag) {
+			bi_errorf("can't use -e, -l, -n, -r with -s (-e -)");
+			return 1;
+		}
+
+		if (!first && (first = *wp)) {
+			wp++;
+			if (*first && (p = strchr(first + 1, '='))) {
+				pat = first;
+				*p++ = '\0';
 				rep = p;
-				*rep++ = '\0';
-			} else
-				hp = histget(id, FALSE);
+				first = *wp++;
+			}
 		}
-
-		if (hp == NULL || hp < history) {
-			/* XXX improve error */
-			bi_errorf("cannot find history");
+		if (last || *wp) {
+			bi_errorf("too many arguments");
 			return 1;
 		}
-		if (pat == NULL) {
-			int len = strlen(*hp);
-			p = strnsave(*hp, len + 1, source->areap);
-			p[len] = '\n';
-			p[len + 1] = '\0';
-		} else if (!(p = histrpl(*hp, pat, rep, gflag)))
+		histbackup(); /* before hist_get_newest() */
+		hp = first ? hist_get(first, FALSE) : hist_get_newest(TRUE);
+		if (!hp)
 			return 1;
-		histbackup();
-		histsave(source->line+1, p, 1);
-		histpush--;
-		return 0;
+		return hist_replace(hp, pat, rep, gflag);
 	}
 
-	if (*wp != NULL) {
-		hbeg = histget(*wp++, TRUE); /* first */
-		if (*wp != NULL)
-			hend = histget(*wp++, TRUE); /* last */
-		else if (lflag)
-			hend = (histptr - hbeg > 16) ? hbeg + 16 : histptr;
-		else
-			hend = hbeg;
-	} else {
-		if (lflag)
-			hbeg = histptr - 16, hend = histptr;
-		else
-			hbeg = hend = histptr - 1;
-		if (hbeg < history)
-			hbeg = history;
-	}
-	if (hbeg == NULL || hend == NULL) {
-		/* XXX improve error */
-		bi_errorf("can't find history");
+	if (editor && (lflag || nflag)) {
+		bi_errorf("can't use -l, -n with -e");
 		return 1;
 	}
 
-	if (lflag)
-		shf = shf_fdopen(1, SHF_WR, (struct shf *) 0);
-	else {
-		nflag++;
-		tf = maketemp(ATEMP);
-		tf->next = e->temps; e->temps = tf;
-		shf = shf_open(tf->name, O_WRONLY|O_CREAT|O_TRUNC, 0666, 0);
-		if (shf == NULL) {
-			bi_errorf("cannot create temp file %s", tf->name);
+	if (!first && (first = *wp))
+		wp++;
+	if (!last && (last = *wp))
+		wp++;
+	if (*wp) {
+		bi_errorf("too many arguments");
+		return 1;
+	}
+	if (!lflag)
+		histbackup(); /* before hist_get_newest() */
+	if (!first) {
+		hfirst = lflag ? hist_get("-16", TRUE) : hist_get_newest(TRUE);
+		if (!hfirst)
 			return 1;
-		}
+		/* can't fail if hfirst did fail */
+		hlast = hist_get_newest(TRUE);
+	} else {
+		/* POSIX says not an error if first/last out of bounds
+		 * when range is specified; at&t ksh and pdksh allow out of
+		 * bounds for -l as well.
+		 */
+		hfirst = hist_get(first, (lflag || last) ? TRUE : FALSE);
+		if (!hfirst)
+			return 1;
+		hlast = last ? hist_get(last, TRUE)
+			    : (lflag ? hist_get_newest(TRUE) : hfirst);
+		if (!hlast)
+			return 1;
+	}
+	if (hfirst > hlast) {
+		char **temp;
+
+		temp = hfirst; hfirst = hlast; hlast = temp;
+		rflag = 1; /* POSIX (doesn't seem to say rflag = !rflag) */
 	}
 
-	for (hp = (rflag ? hend : hbeg); rflag ? (hp >= hbeg) : (hp <= hend);
-	      rflag ? hp-- : hp++) {
-		if (lflag) {
-			char *s, *t;
+	/* List history */
+	if (lflag) {
+		char *s, *t;
+		char *nfmt = nflag ? "\t" : "%d\t";
 
-			if (!nflag)
-				shf_fprintf(shf, "%d",
-					source->line - (int)(histptr-hp));
-			shf_putchar('\t', shf);
+		for (hp = rflag ? hlast : hfirst;
+		     hp >= hfirst && hp <= hlast; hp += rflag ? -1 : 1)
+		{
+			shf_fprintf(shl_stdout, nfmt,
+				hist_source->line - (int) (histptr - hp));
 			/* print multi-line commands correctly */
 			for (s = *hp; (t = strchr(s, '\n')); s = t)
-				shf_fprintf(shf, "%.*s\t", ++t - s, s, null);
-			shf_fprintf(shf, "%s\n", s);
-		} else
-			shf_fprintf(shf, "%s\n", *hp);
+				shf_fprintf(shl_stdout, "%.*s\t", ++t - s, s);
+			shf_fprintf(shl_stdout, "%s\n", s);
+		}
+		shf_flush(shl_stdout);
+		return 0;
 	}
 
-	shf_flush(shf);
-	if (lflag)
-		return 0;
-	shf_close(shf);
+	/* Run editor on selected lines, then run resulting commands */
+
+	tf = maketemp(ATEMP);
+	tf->next = e->temps; e->temps = tf;
+	shf = shf_open(tf->name, O_WRONLY|O_CREAT|O_TRUNC, 0666, 0);
+	if (!shf) {
+		bi_errorf("cannot create temp file %s - %s",
+			tf->name, strerror(errno));
+		return 1;
+	}
+	for (hp = rflag ? hlast : hfirst;
+	     hp >= hfirst && hp <= hlast; hp += rflag ? -1 : 1)
+		shf_fprintf(shf, "%s\n", *hp);
+	if (shf_close(shf) == EOF) {
+		bi_errorf("error writing temporary file - %s", strerror(errno));
+		return 1;
+	}
 
 	setstr(local("_"), tf->name);
 
 	/* XXX: source should not get trashed by this.. */
 	{
 		Source *sold = source;
+		int ret;
 
-		if (cmd)
-			command(cmd); /* edit temp file */
-		else
-			command("${FCEDIT:-/bin/ed} $_");
+		ret = command(editor ? editor : "${FCEDIT:-/bin/ed} $_");
 		source = sold;
+		if (ret)
+			return ret;
 	}
 
 	{
 		struct stat statb;
 		XString xs;
-		char *xp, *p;
+		char *xp;
 		int n;
-		int first = 1;
 
-		if (!(shf = shf_open(tf->name, O_RDONLY, 0666, 0))) {
+		if (!(shf = shf_open(tf->name, O_RDONLY, 0, 0))) {
 			bi_errorf("cannot open temp file %s", tf->name);
 			return 1;
 		}
 
-		histbackup();
 		n = fstat(shf_fileno(shf), &statb) < 0 ? 128
 			: statb.st_size + 1;
-		Xinit(xs, xp, n, source->areap);
-		while (1) {
-			/* read a line */
-			xp = Xstring(xs, xp);
-			while ((p = shf_getse(xp, Xnleft(xs, xp), shf))) {
-				xp = p;
-				if (Xstring(xs, xp) != xp && p[-1] == '\n')
-					break;
+		Xinit(xs, xp, n, hist_source->areap);
+		while ((n = shf_read(xp, Xnleft(xs, xp), shf)) > 0) {
+			xp += n;
+			if (Xnleft(xs, xp) <= 0)
 				XcheckN(xs, xp, Xlength(xs, xp));
-			}
-			if (shf_error(shf)) {
-				bi_errorf("error reading temp file %s - %s",
-					tf->name, strerror(shf_errno(shf)));
-				shf_close(shf);
-				return 1;
-			}
-			if (!p && xp == Xstring(xs, xp))
-				break;
-			strip_nuls(Xstring(xs, xp), Xlength(xs, xp));
-# ifdef EASY_HISTORY
-			if (!first) {
-				histappend(Xstring(xs, xp), 1); 
-				continue;
-			}
-			first = 0;
-# endif
-			histpush--;
-			histsave(source->line+1, Xstring(xs, xp), 1);
+		}
+		if (n < 0) {
+			bi_errorf("error reading temp file %s - %s",
+				tf->name, strerror(shf_errno(shf)));
+			shf_close(shf);
+			return 1;
 		}
 		shf_close(shf);
+		*xp = '\0';
+		strip_nuls(Xstring(xs, xp), Xlength(xs, xp));
+		return hist_execute(Xstring(xs, xp));
+	}
+}
+
+/* Save cmd in history, execute cmd (cmd gets trashed) */
+static int
+hist_execute(cmd)
+	char *cmd;
+{
+	Source *sold;
+	int ret;
+	char *p, *q;
+
+	for (p = cmd; p; p = q) {
+		if ((q = strchr(p, '\n'))) {
+			*q++ = '\0'; /* kill the newline */
+			if (!*q) /* ignore trailing newline */
+				q = (char *) 0;
+		}
+#ifdef EASY_HISTORY
+		if (p != cmd)
+			histappend(p, TRUE);
+		else
+#endif /* EASY_HISTORY */
+			histsave(++(hist_source->line), p, 1);
+
+		shellf("%s\n", p); /* POSIX doesn't say this is done... */
+		if ((p = q)) /* restore \n (trailing \n not restored) */
+			q[-1] = '\n';
 	}
 
-	return 0;
+	/* Commands are executed here instead of pushing them onto the
+	 * input 'cause posix says the redirection and variable assignments
+	 * in
+	 *	X=y fc -e - 42 2> /dev/null
+	 * are to effect the repeated commands environment.
+	 */
+	/* XXX: source should not get trashed by this.. */
+	sold = source;
+	ret = command(cmd);
+	source = sold;
+	return ret;
+}
+
+static int
+hist_replace(hp, pat, rep, global)
+	char **hp;
+	char *pat;
+	char *rep;
+	int global;
+{
+	char *line;
+
+	if (!pat)
+		line = strsave(*hp, ATEMP);
+	else {
+		char *s, *s1, *last = (char *) 0;
+		int pat_len = strlen(pat);
+		int rep_len = strlen(rep);
+		int len;
+		XString xs;
+		char *xp;
+
+		Xinit(xs, xp, 128, ATEMP);
+		for (s = *hp; (s1 = strstr(s, pat)); s = s1 + pat_len) {
+			len = s1 - s;
+			XcheckN(xs, xp, len + rep_len);
+			memcpy(xp, s, len);		/* first part */
+			xp += len;
+			memcpy(xp, rep, rep_len);	/* replacement */
+			xp += rep_len;
+			if (!global)
+				break;
+		}
+		if (xp == Xstring(xs, xp)) {
+			bi_errorf("substitution failed");
+			return 1;
+		}
+		len = strlen(s) + 1;
+		XcheckN(xs, xp, len);
+		memcpy(xp, last, len);
+		xp += len;
+		line = Xclose(xs, xp);
+	}
+	return hist_execute(line);
+}
+
+/*
+ * get pointer to history given pattern
+ * pattern is a number or string
+ */
+static char **
+hist_get(str, approx)
+	char *str;
+	int approx;
+{
+	char **hp = (char **) 0;
+	int n;
+
+	if (getn(str, &n)) {
+		/* n + 1 if n < 0 to account for the histbackup() */
+		hp = histptr + (n < 0 ? n + 1 : (n - hist_source->line));
+		if (hp < history) {
+			if (approx)
+				hp = hist_get_oldest(TRUE);
+			else {
+				bi_errorf("%s: not in history", str);
+				hp = (char **) 0;
+			}
+		} else if (hp > histptr) {
+			if (approx)
+				hp = hist_get_newest(TRUE);
+			else {
+				bi_errorf("%s: not in history", str);
+				hp = (char **) 0;
+			}
+		}
+	} else {
+		int anchored = *str == '?' ? (++str, 0) : 1;
+
+		n = findhist(histptr - history, 0, str, anchored);
+		if (n < 0) {
+			bi_errorf("%s: not in history", str);
+			hp = (char **) 0;
+		} else
+			hp = &history[n];
+	}
+	return hp;
+}
+
+/* Return a pointer to the newest command in the history */
+static char **
+hist_get_newest(print_err)
+	int print_err;
+{
+	if (histptr < history) {
+		if (print_err)
+			bi_errorf("no history (yet)");
+		return (char **) 0;
+	}
+	return histptr;
+}
+
+/* Return a pointer to the newest command in the history */
+static char **
+hist_get_oldest(print_err)
+	int print_err;
+{
+	if (histptr < history) {
+		if (print_err)
+			bi_errorf("no history (yet)");
+		return (char **) 0;
+	}
+	return history;
 }
 
 /******************************/
@@ -268,82 +441,14 @@ histbackup()
 {
 	static int last_line = -1;
 
-	if (histptr > history && last_line != source->line) { 
-		source->line--;
+	if (histptr > history && last_line != hist_source->line) { 
+		hist_source->line--;
 		afree((void*)*histptr, APERM);
 		histptr--;
-		last_line = source->line;
+		last_line = hist_source->line;
 	}
 }
 
-/*
- * get pointer to history given pattern
- * pattern is a number or string
- */
-static char **
-histget(str, approx)
-	char *str;
-	int approx;
-{
-	register char **hp = NULL;
-	int n;
-
-	if (getn(str, &n)) {
-		if (n < 0) {
-			hp = histptr + n;
-			if (approx && hp < history)
-				hp = history;
-		} else
-			hp = histptr + (n - source->line);
-	} else {
-		int anchored = *str == '?' ? (++str, 0) : 1;
-
-		hp = &history[findhist((histptr - 1) - history, 0, str,
-					anchored)];
-	}
-
-	return (history <= hp && hp <= histptr) ? hp : NULL;
-}
-
-static char *
-histrpl(s, pat, rep, global)
-	char *s;
-	char *pat, *rep;
-	int global;
-{
-	char *s1, *last = NULL;
-	int pat_len = strlen(pat);
-	int rep_len = strlen(rep);
-	XString xs;
-	char *xp;
-
-	Xinit(xs, xp, 128, source->areap);
-	while ((s1 = strstr(s, pat))) {
-		XcheckN(xs, xp, (s1 - s) + rep_len);
-		memcpy(xp, s, s1 - s);		/* first part */
-		xp += s1 - s;
-		memcpy(xp, rep, rep_len);	/* replacement */
-		xp += rep_len;
-		s = last = s1 + pat_len;
-		if (!global)
-			break;
-	}
-	if (last) {
-		int last_len = strlen(last);
-
-		XcheckN(xs, xp, last_len);
-		memcpy(xp, last, last_len);
-		xp += last_len;
-		*xp++ = '\n';
-		*xp++ = '\0';
-	} else {
-		bi_errorf("substitution failed");
-		Xfree(xs, xp);
-		return (char *) 0;
-	}
-	return Xclose(xs, xp);
-}
-
 /*
  * Return the current position.
  */
@@ -377,7 +482,7 @@ histnum(n)
 }
 
 /*
- * This will become unecessary if histget is modified to allow
+ * This will become unecessary if hist_get is modified to allow
  * searching from positions other than the end, and in either 
  * direction.
  */
@@ -464,11 +569,11 @@ sethistfile(name)
 		hname = NULL;
 		/* let's reset the history */
 		histptr = history - 1;
-		source->line = 0;
+		hist_source->line = 0;
 	}
 # endif
 
-	hist_init(source);
+	hist_init(hist_source);
 }
 
 /*
@@ -555,6 +660,8 @@ hist_init(s)
 
 	hstarted = 1;
 
+	hist_source = s;
+
 	if ((f = strval(global("HISTFILE"))) == NULL || *f == '\0') {
 # if 1 /* Don't use history file unless the user asks for it */
 		hname = NULL;
@@ -590,7 +697,7 @@ hist_init(s)
 			if (contin)
 				histappend(hline + pos, 0);
 			else {
-				s->line++;
+				hist_source->line++;
 				histsave(0, hline + pos, 0);
 			}
 			pos = end - hline + 1;
@@ -709,6 +816,8 @@ hist_init(s)
 
 	hstarted = 1;
 	
+	hist_source = s;
+
 	hname = strval(global("HISTFILE"));
 	if (hname == NULL)
 		return;
@@ -723,7 +832,7 @@ hist_init(s)
 
 	(void) flock(histfd, LOCK_EX);
 
-	hsize = lseek(histfd, 0L, L_XTND);
+	hsize = lseek(histfd, 0L, SEEK_END);
 
 	if (hsize == 0) {
 		/* add magic */
@@ -758,11 +867,11 @@ hist_init(s)
 				goto retry;
 			}
 		}
-		histload(s, base+2, hsize-2);
+		histload(hist_source, base+2, hsize-2);
 		munmap((caddr_t)base, hsize);
 	}
 	(void) flock(histfd, LOCK_UN);
-	hsize = lseek(histfd, 0L, L_XTND);
+	hsize = lseek(histfd, 0L, SEEK_END);
 }
 
 typedef enum state {
@@ -978,7 +1087,7 @@ writehistfile(lno, cmd)
 	char	hdr[5];
 	
 	(void) flock(histfd, LOCK_EX);
-	sizenow = lseek(histfd, 0L, L_XTND);
+	sizenow = lseek(histfd, 0L, SEEK_END);
 	if (sizenow != hsize) {
 		/*
 		 *	Things have changed
@@ -994,10 +1103,10 @@ writehistfile(lno, cmd)
 				munmap((caddr_t)base, sizenow);
 				goto bad;
 			}
-			source->line--;
-			histload(source, new, bytes);
-			source->line++;
-			lno = source->line;
+			hist_source->line--;
+			histload(hist_source, new, bytes);
+			hist_source->line++;
+			lno = hist_source->line;
 			munmap((caddr_t)base, sizenow);
 			hsize = sizenow;
 		} else {
@@ -1017,7 +1126,7 @@ writehistfile(lno, cmd)
 	hdr[4] = lno&0xff;
 	(void) write(histfd, hdr, 5);
 	(void) write(histfd, cmd, strlen(cmd)+1);
-	hsize = lseek(histfd, 0L, L_XTND);
+	hsize = lseek(histfd, 0L, SEEK_END);
 	(void) flock(histfd, LOCK_UN);
 	return;
 bad:
