@@ -53,6 +53,7 @@ warningf(fileline, fmt, va_alist)
 	SH_VA_START(va, fmt);
 	shf_vfprintf(shl_out, fmt, va);
 	va_end(va);
+	shf_putchar('\n', shl_out);
 	shf_flush(shl_out);
 }
 
@@ -414,7 +415,8 @@ maketemp(ap)
 	len = strlen(tmp) + 3 + 20 + 20 + 1;
 	tp = (struct temp *) alloc(sizeof(struct temp) + len, ap);
 	tp->name = path = (char *) &tp[1];
-	shf_snprintf(path, len, "%s/sh%05u%02u",
+	/* Note that temp files need to fit 8.3 DOS limits */
+	shf_snprintf(path, len, "%s/sh%05u.%03x",
 		tmp, (unsigned) procpid, inc++);
 	close(creat(path, 0600));	/* to get safe permissions */
 	tp->next = NULL;
