@@ -187,7 +187,7 @@ static struct shf	*shl_j;
 /* On some systems, the kernel doesn't count zombie processes when checking
  * if a process group is valid, which can cause problems in creating the
  * pipeline "cmd1 | cmd2": if cmd1 can die (and go into the zombie state)
- * before cmd2 is started, the kernel doesn't allow the setpgrp() for cmd2
+ * before cmd2 is started, the kernel doesn't allow the setpgid() for cmd2
  * to succeed.  Solution is to create a pipe between the parent and the first
  * process; the first process doesn't do anything until the pipe is closed
  * and the parent doesn't close the pipe until all the processes are started.
@@ -272,20 +272,27 @@ j_init(mflagset)
 void
 j_exit()
 {
-#ifdef JOBS
-	/* kill stopped jobs */
+	/* kill stopped, and possibly running, jobs */
 	Job	*j;
 	int	killed = 0;
 
-	for (j = job_list; j != (Job *) 0; j = j->next)
+	for (j = job_list; j != (Job *) 0; j = j->next) {
+#ifdef JOBS
 		if (j->ppid == procpid && j->state == PSTOPPED) {
 			killed = 1;
 			killpg(j->pgrp, SIGCONT);
 			killpg(j->pgrp, SIGHUP);
 		}
+#endif /* JOBS */
+		if (Flag(FLOGIN) && !Flag(FNOHUP) && procpid == kshpid
+		    && j->ppid == procpid && j->state == PRUNNING)
+		{
+			killed = 1;
+			killpg(j->pgrp, SIGHUP);
+		}
+	}
 	if (killed)
 		sleep(1);
-#endif /* JOBS */
 	j_notify();
 
 #ifdef JOBS
@@ -325,7 +332,7 @@ j_change()
 		ttypgrp_ok = tty_fd >= 0 && tty_devtty;
 
 		if (ttypgrp_ok && (our_pgrp = getpgID()) < 0) {
-			warningf(FALSE, "j_init: getpgrp() failed: %s\n",
+			warningf(FALSE, "j_init: getpgrp() failed: %s",
 				strerror(errno));
 			ttypgrp_ok = 0;
 		}
@@ -338,7 +345,7 @@ j_change()
 
 				if ((ttypgrp = tcgetpgrp(tty_fd)) < 0) {
 					warningf(FALSE,
-					"j_init: tcgetpgrp() failed: %s\n",
+					"j_init: tcgetpgrp() failed: %s",
 						strerror(errno));
 					ttypgrp_ok = 0;
 					break;
@@ -354,13 +361,13 @@ j_change()
 		if (ttypgrp_ok && our_pgrp != kshpid) {
 			if (setpgid(0, kshpid) < 0) {
 				warningf(FALSE,
-					"j_init: setpgid() failed: %s\n",
+					"j_init: setpgid() failed: %s",
 					strerror(errno));
 				ttypgrp_ok = 0;
 			} else {
 				if (tcsetpgrp(tty_fd, kshpid) < 0) {
 					warningf(FALSE,
-					"j_init: tcsetpgrp() failed: %s\n",
+					"j_init: tcsetpgrp() failed: %s",
 						strerror(errno));
 					ttypgrp_ok = 0;
 				} else
@@ -374,13 +381,12 @@ j_change()
 
 			if (ioctl(tty_fd, TIOCSETD, &ldisc) < 0)
 				warningf(FALSE,
-				"j_init: can't set new line discipline: %s\n",
+				"j_init: can't set new line discipline: %s",
 					strerror(errno));
 		}
 #  endif /* NTTYDISC && TIOCSETD */
 		if (!ttypgrp_ok)
-			warningf(FALSE,
-				"warning: won't have full job control\n");
+			warningf(FALSE, "warning: won't have full job control");
 # endif /* TTY_PGRP */
 		if (tty_fd >= 0)
 			get_tty(tty_fd, &tty_state);
@@ -605,6 +611,8 @@ exchild(t, flags, close_fd)
 		tty_close();
 		cleartraps();
 		execute(t, flags|XEXEC); /* no return */
+		internal_errorf(0, "exchild: execute() returned");
+		unwind(LLEAVE);
 		/* NOTREACHED */
 	}
 
@@ -674,7 +682,7 @@ waitlast()
 	j = last_job;
 	if (!j || !(j->flags & JF_STARTED)) {
 		if (!j)
-			warningf(TRUE, "waitlast: no last job\n");
+			warningf(TRUE, "waitlast: no last job");
 		else
 			internal_errorf(0, "waitlast: not started");
 #ifdef JOB_SIGS
@@ -885,7 +893,7 @@ j_resume(cp, bg)
 				set_tty(tty_fd, &tty_state, TF_NONE);
 			if (ttypgrp_ok && tcsetpgrp(tty_fd, our_pgrp) < 0) {
 				warningf(TRUE,
-				"fg: 2nd tcsetpgrp(%d, %d) failed: %s\n",
+				"fg: 2nd tcsetpgrp(%d, %d) failed: %s",
 					tty_fd, our_pgrp,
 					strerror(errno));
 			}
@@ -909,6 +917,33 @@ j_resume(cp, bg)
 }
 #endif /* JOBS */
 
+/* are there any running or stopped jobs ? */
+int
+j_stopped_running()
+{
+	Job	*j;
+	int	which = 0;
+
+	for (j = job_list; j != (Job *) 0; j = j->next) {
+#ifdef JOBS
+		if (j->ppid == procpid && j->state == PSTOPPED)
+			which |= 1;
+#endif /* JOBS */
+		if (Flag(FLOGIN) && !Flag(FNOHUP) && procpid == kshpid
+		    && j->ppid == procpid && j->state == PRUNNING)
+			which |= 2;
+	}
+	if (which) {
+		shellf("You have %s%s%s jobs\n",
+			which & 1 ? "stopped" : "",
+			which == 3 ? " and " : "",
+			which & 2 ? "running" : "");
+		return 1;
+	}
+
+	return 0;
+}
+
 #ifdef JOBS
 /* are there any stopped jobs ? */
 int
@@ -1152,7 +1187,7 @@ j_waitj(j, flags, where)
 		if (Flag(FMONITOR) && ttypgrp_ok && j->pgrp) {
 			if (tcsetpgrp(tty_fd, our_pgrp) < 0) {
 				warningf(TRUE,
-				"j_waitj: tcsetpgrp(%d, %d) failed: %s\n",
+				"j_waitj: tcsetpgrp(%d, %d) failed: %s",
 					tty_fd, our_pgrp,
 					strerror(errno));
 			}
@@ -1232,7 +1267,7 @@ j_sigchld(sig)
 	/* Don't wait for any processes if a job is partially started.
 	 * This is so we don't do away with the process group leader
 	 * before all the processes in a pipe line are started (so the
-	 * setpgrp() won't fail)
+	 * setpgid() won't fail)
 	 */
 	for (j = job_list; j; j = j->next)
 		if (j->ppid == procpid && !(j->flags & JF_STARTED)) {
@@ -1262,7 +1297,7 @@ j_sigchld(sig)
 found:
 		if (j == (Job *) 0) {
 			/* Can occur if process has kids, then execs shell
-			warningf(TRUE, "bad process waited for (pid = %d)\n",
+			warningf(TRUE, "bad process waited for (pid = %d)",
 				pid);
 			 */
 			t0 = t1;
