@@ -691,17 +691,9 @@ getsc_()
 			return 0;
 
 		  case STTY:
-#ifdef HISTORY
-			if (histpush < 0) {	/* commands pushed by dofc */
-				s->type = SHIST;
-				s->str = null;
-				continue;
-			}
-#endif /* HISTORY */
 			s->str = line;
 			line[0] = '\0';
 			mprint(); /* print mail messages */
-			pprompt(prompt);
 #ifdef KSH
 			if (ksh_tmout) {
 				ksh_tmout_state = TMOUT_READING;
@@ -721,11 +713,12 @@ getsc_()
 			else
 #endif
 			{
+				pprompt(prompt, 0);
 				/*
-				 * This allows the arival of a SIGCHLD 
+				 * This allows the arrival of a SIGCHLD 
 				 * to not disturb us until we are ready. 
 				 */
-				while ((c = read(0, line, LINE)) < 0
+				while ((c = blocking_read(0, line, LINE)) < 0
 					&& errno == EINTR)
 					if (trap)
 						runtraps(0);
@@ -759,7 +752,7 @@ getsc_()
 					if (cur_prompt == PS2)
 						histappend(line, 1);
 					else
-# endif
+# endif /* EASY_HISTORY */
 					    {
 						s->line++;
 						histsave(s->line, s->str, 1);
@@ -767,27 +760,9 @@ getsc_()
 				}
 #endif /* HISTORY */
 			}
-			set_prompt(PS2);
+			set_prompt(PS2, (Source *) 0);
 			break;
 
-#ifdef HISTORY
-		  case SHIST:
-			if (histpush == 0) {
-				s->type = STTY;
-				s->str = null;
-				continue;
-			}
-			s->str = histptr[++histpush];
-			s->line++;
-			shellf("%s\n", s->str);
-			strcpy(line, s->str);
-			s->str = strchr(line, '\0');
-			*s->str++ = '\n';
-			*s->str = '\0';
-			s->str = line;
-			break;
-#endif /* HISTORY */
-
 		  case SFILE:
 		  case SSTDIN:
 		  {
@@ -876,7 +851,7 @@ getsc_()
 			return 0;
 		}
 		if (s->flags & SF_ECHO) {
-			shf_write(s->str, strlen(s->str), shl_out);
+			shf_puts(s->str, shl_out);
 			shf_flush(shl_out);
 		}
 	}
@@ -884,14 +859,45 @@ getsc_()
 }
 
 void
-set_prompt(to)
+set_prompt(to, s)
 	int to;
+	Source *s;
 {
 	cur_prompt = to;
 
 	switch (to) {
 	case PS1: /* command */
-		prompt = substitute(strval(global("PS1")), 0);
+		/* Substitute ! and !! here, before substitutions are done
+		 * so ! in expanded variables are not expanded.
+		 * NOTE: this is not what at&t ksh does (it does it after
+		 * substitutions, POSIX doesn't say which is to be done.
+		 */
+		{
+			struct shf *shf;
+			char *ps1;
+			Area *saved_atemp;
+
+			ps1 = strval(global("PS1"));
+			shf = shf_sopen((char *) 0, strlen(ps1),
+				SHF_WR | SHF_DYNAMIC, (struct shf *) 0);
+			while (*ps1) {
+				if (*ps1 != '!' || *++ps1 == '!')
+					shf_putchar(*ps1++, shf);
+				else
+					shf_fprintf(shf, "%d",
+						s ? s->line + 1 : 0);
+			}
+			ps1 = shf_sclose(shf);
+			saved_atemp = ATEMP;
+			newenv(E_ERRH);
+			if (setjmp(e->jbuf)) {
+				prompt = safe_prompt;
+				warningf(TRUE, "error during expansion of PS1");
+			} else
+				prompt = strsave(substitute(ps1, 0),
+						 saved_atemp);
+			quitenv();
+		}
 		break;
 
 	case PS2: /* command continuation */
@@ -902,18 +908,50 @@ set_prompt(to)
 
 /* See also related routine, promptlen() in edit.c */
 void
-pprompt(cp)
+pprompt(cp, ntruncate)
 	register char *cp;
+	int ntruncate;
 {
-	while (*cp != 0)
+#if 0
+	char nbuf[32];
+	int c;
+
+	while (*cp != 0) {
 		if (*cp != '!')
-			shf_putc(*cp++, shl_out);
+			c = *cp++;
+		else if (*++cp == '!')
+			c = *cp++;
+		else {
+			int len;
+			char *p;
+
+			shf_snprintf(p = nbuf, sizeof(nbuf), "%d",
+				source->line + 1);
+			len = strlen(nbuf);
+			if (ntruncate) {
+				if (ntruncate >= len) {
+					ntruncate -= len;
+					continue;
+				}
+				p += ntruncate;
+				len -= ntruncate;
+				ntruncate = 0;
+			}
+			shf_write(p, len, shl_out);
+			continue;
+		}
+		if (ntruncate)
+			--ntruncate;
 		else
-			if (*++cp == '!')
-				shf_putc(*cp++, shl_out);
-			else
-				shf_fprintf(shl_out, "%d", source->line + 1);
-	shf_flush(shl_out);
+			shf_putc(c, shl_out);
+	}
+#endif /* 0 */
+	if (ntruncate)
+		shellf("%.*s", ntruncate, cp);
+	else {
+		shf_puts(cp, shl_out);
+		shf_flush(shl_out);
+	}
 }
 
 /* Read the variable part of a ${...} expression (ie, up to but not including
