@@ -6,6 +6,7 @@
 #include "ksh_stat.h"
 #include "ksh_dir.h"
 
+
 #ifndef HAVE_MEMSET
 void *
 memset(d, c, n)
@@ -22,7 +23,6 @@ memset(d, c, n)
 }
 #endif /* !HAVE_MEMSET */
 
-
 #if !defined(HAVE_MEMMOVE) && !defined(HAVE_BCOPY)
 void *
 memmove(d, s, n)
@@ -189,41 +189,6 @@ strerror(err)
 #endif /* !HAVE_STRERROR */
 
 
-#ifndef HAVE_GETCWD
-# include <sys/param.h>
-# ifndef MAXPATHLEN
-#  define MAXPATHLEN 1024
-# endif /* MAXPATHLEN */
-
-char *
-getcwd(buf, size)
-	char	*buf;
-	int	size;
-{
-	extern char *getwd ARGS((char *));
-	char tmp_buf[MAXPATHLEN + 1];
-
-	if (size <= 0) {
-		errno = EINVAL;
-		return (char *) 0;
-	}
-	if (getwd(size < MAXPATHLEN ? tmp_buf : buf) == (char *) 0) {
-		errno = EACCES;
-		return (char *) 0;
-	}
-	if (size < MAXPATHLEN) {
-		tmp_buf[sizeof(tmp_buf) - 1] = '\0';
-		if (strlen(tmp_buf) >= size) {
-			errno = ERANGE;
-			return (char *) 0;
-		}
-		strcpy(buf, tmp_buf);
-	}
-	return buf;
-}
-#endif /* !HAVE_GETCWD */
-
-
 #ifdef TIMES_BROKEN
 # include "ksh_time.h"
 # include "ksh_times.h"
