@@ -50,10 +50,11 @@
 
 /* Define COMPLEX_HISTORY if you want at&t ksh style history files (ie, file
  * is updated after each command is read; concurrent ksh's read each other's
- * commands, etc.).  This option uses the mmap() function - if mmap isn't
- * available, the option is automatically undefined.  If option is not defined,
- * a simplier history mechanism which reads/saves the history at startup/exit
- * time, respectively, is used.  COMPLEX_HISTORY is courtesy of Peter Collinson.
+ * commands, etc.). This option uses the mmap() and flock() functions - if
+ * these aren't available, the option is automatically undefined.  If this
+ * option is not defined, a simplier history mechanism which reads/saves the
+ * history at startup/exit time, respectively, is used.  COMPLEX_HISTORY is
+ * courtesy of Peter Collinson. 
  */
 #undef COMPLEX_HISTORY
 
@@ -67,6 +68,16 @@
  */
 #undef POSIXLY_CORRECT
 
+/* Define DEFAULT_ENV to be the name of the file (eg, "/etc/default.env") to
+ * include if the ENV environment variable is not set when the shell starts
+ * up.  This can be useful when used with rsh(1) which creates a non-login
+ * shell (ie, profile not read) with an empty environment (ie, ENV not set).
+ * Setting ENV to null disables the inclusion of DEFAULT_ENV.
+ * NOTE: this is a non-standard feature (ie, at&t ksh has no default
+ * environment) - undefining this disables the use of a default ENV file.
+ */
+#undef DEFAULT_ENV
+
 /* Define SWTCH to handle SWITCH character, for use with shell layers (shl(1)).
  * This has not been tested for some time.
  */
