@@ -60,7 +60,7 @@ bool_t 	x_mode		ARGS((bool_t onoff));
 int 	promptlen	ARGS((char *cp));
 void	set_editmode	ARGS((char *ed));
 /* emacs.c: most prototypes in edit.h */
-int 	x_bind		ARGS((char *a1, char *a2, int macro));
+int 	x_bind		ARGS((char *a1, char *a2, int macro, int list));
 /* eval.c */
 char *	substitute	ARGS((const char *cp, int f));
 char **	eval		ARGS((char **ap, int f));
@@ -79,8 +79,8 @@ struct tbl *	findcom	ARGS((char *name, int flags));
 void 	flushcom	ARGS((int all));
 char *	search		ARGS((char *name, char *path, int mode));
 /* expr.c */
-long 	evaluate	ARGS((const char *expr));
-void	v_evaluate	ARGS((struct tbl *vp, const char *expr));
+int 	evaluate	ARGS((const char *expr, long *rval, int error_ok));
+int	v_evaluate	ARGS((struct tbl *vp, const char *expr, volatile int error_ok));
 /* history.c */
 void	init_histvec	ARGS((void));
 void 	hist_init	ARGS((Source *s));
@@ -140,14 +140,14 @@ int 	j_resume	ARGS((char *cp, int bg));
 int 	j_jobs		ARGS((char *cp, int slp, int nflag));
 void 	j_notify	ARGS((void));
 pid_t	j_async		ARGS((void));
-int 	j_stopped	ARGS((void));
+int 	j_stopped_running	ARGS((void));
 /* lex.c */
 int 	yylex		ARGS((int cf));
 void 	yyerror		ARGS((const char *fmt, ...))
 				GCC_FUNC_ATTR2(noreturn, format(printf, 1, 2));
 Source * pushs		ARGS((int type, Area *areap));
-void	set_prompt	ARGS((int to));
-void 	pprompt		ARGS((char *cp));
+void	set_prompt	ARGS((int to, Source *s));
+void 	pprompt		ARGS((char *cp, int ntruncate));
 /* mail.c */
 void 	mcheck		ARGS((void));
 void 	mbset		ARGS((char *p));
@@ -177,7 +177,7 @@ int	parse_args	ARGS((char **argv, int what, int *setargsp));
 int 	getn		ARGS((char *as, int *ai));
 int 	bi_getn		ARGS((char *as, int *ai));
 char *	strerror	ARGS((int i));
-int 	gmatch		ARGS((char *s, char *p));
+int 	gmatch		ARGS((char *s, char *p, int isfile));
 void 	qsortp		ARGS((void **base, size_t n, int (*f)()));
 int 	xstrcmp		ARGS((void *p1, void *p2));
 void	ksh_getopt_reset ARGS((Getopt *go, int));
@@ -188,10 +188,15 @@ void	print_columns	ARGS((struct shf *shf, int n,
 			      void *arg, int max_width));
 int	strip_nuls	ARGS((char *buf, int nbytes));
 char	*str_zcpy	ARGS((char *dst, char *src, int dsize));
+int	blocking_read	ARGS((int fd, char *buf, int nbytes));
+int	reset_nonblock	ARGS((int fd));
+char	*ksh_get_wd	ARGS((char *buf, int bsize));
 /* path.c */
-int	make_path	ARGS((char *curpath, char *file, char **pathlist,
-				char *result, int len));
+int	make_path	ARGS((char *cwd, char *file, char **pathlist,
+				XString *xsp, int *phys_pathp));
 void	simplify_path	ARGS((char *path));
+char	*get_phys_path	ARGS((char *path));
+void	set_current_wd	ARGS((char *path));
 /* syn.c */
 void 	initkeywords	ARGS((void));
 struct op * compile	ARGS((Source *s));
@@ -242,14 +247,14 @@ struct tbl *	strint	ARGS((struct tbl *vq, struct tbl *vp));
 void 	setint		ARGS((struct tbl *vq, long n));
 int	getint		ARGS((struct tbl *vp, long *nump));
 struct tbl *	typeset	ARGS((char *var, int set, int clr, int field, int base));
-void 	unset		ARGS((struct tbl *vp));
+void 	unset		ARGS((struct tbl *vp, int array_ref));
 char  * skip_varname	ARGS((const char *s, int aok));
 char	*skip_wdvarname ARGS((char *s, int aok));
 int	is_wdvarname	ARGS((char *s, int aok));
 int	is_wdvarassign	ARGS((char *s));
 char **	makenv		ARGS((void));
 int	array_ref_len	ARGS((const char *cp));
-char *  basename	ARGS((char *str));
+char *  arrayname	ARGS((char *str));
 void    set_array	ARGS((char *var, int reset, char **vals));
 /* version.c */
 /* vi.c: prototypes in edit.h */
