@@ -65,13 +65,16 @@ int	    shf_fdclose	ARGS((struct shf *shf));
 char	   *shf_sclose	ARGS((struct shf *shf));
 int	    shf_finish	ARGS((struct shf *shf));
 int	    shf_flush	ARGS((struct shf *shf));
+int	    shf_seek	ARGS((struct shf *shf, off_t where, int from));
 int	    shf_read	ARGS((char *buf, int bsize, struct shf *shf));
 char	   *shf_getse	ARGS((char *buf, int bsize, struct shf *shf));
 int	    shf_getchar	ARGS((struct shf *shf));
 int	    shf_ungetc	ARGS((int c, struct shf *shf));
 int	    shf_putchar	ARGS((int c, struct shf *shf));
+int	    shf_puts	ARGS((char *s, struct shf *shf));
 int	    shf_write	ARGS((char *buf, int nbytes, struct shf *shf));
 int	    shf_fprintf ARGS((struct shf *shf, const char *fmt, ...));
 int	    shf_snprintf ARGS((char *buf, int bsize, char *fmt, ...));
+char	    *shf_smprintf ARGS((char *fmt, ...));
 /* stdarg/varargs may not have been included yet.. */
 int	    shf_vfprintf ARGS((/*struct shf *, char *fmt, va_list args*/));
