@@ -218,7 +218,7 @@ sigaction(sig, act, oact)
     nsv.sv_flags = SV_INTERRUPT;	/* punt */
     sigvec(sig, &nsv, &osv);
     oldh = osv.sv_handler;
-# else USE_SIGMASK
+# else /* USE_SIGMASK */
     oldh = signal(sig, act->sa_handler);
 # endif /* USE_SIGMASK */
 #endif
