@@ -10,11 +10,14 @@ set -e
 in=tmpi$$.c
 out=tmpo$$.c
 ecode=1
-trap 'rm -f $in $out; trap 0; exit $ecode' 0 1 2 13 15
+trapsigs='0 1 2 13 15'
+trap 'rm -f $in $out; trap 0; exit $ecode' $trapsigs
 
 CPP="${1-cc -E}"
 
-(echo '#include "sh.h"';
+# The trap here to make up for a bug in bash (1.14.3(1)) that calls the trap
+(trap $trapsigs;
+ echo '#include "sh.h"';
  echo '	{ QwErTy SIGNALS , "DUMMY" , "hook for number of signals" },';
  sed -e '/^[	 ]*#/d' -e 's/^[	 ]*\([^ 	][^ 	]*\)[	 ][	 ]*\(.*[^ 	]\)[ 	]*$/#ifdef SIG\1\
 	{ QwErTy SIG\1 , "\1", "\2" },\
