@@ -3,6 +3,7 @@
  */
 
 #include "sh.h"
+#include "c_test.h"
 
 struct multiline_state {
 	int	on;		/* set in multiline commands (\n becomes ;) */
@@ -28,15 +29,16 @@ static char **	wordlist	ARGS((void));
 static struct op *block		ARGS((int type, struct op *t1, struct op *t2,
 				      char **wp));
 static struct op *newtp		ARGS((int type));
-static void	syntaxerr	ARGS((void)) GCC_FUNC_ATTR(noreturn);
+static void	syntaxerr	ARGS((char *what)) GCC_FUNC_ATTR(noreturn);
 static void	multiline_push ARGS((struct multiline_state *save, int tok));
 static void	multiline_pop ARGS((struct multiline_state *saved));
 static int	assign_command ARGS((char *s));
 #ifdef KSH
-static void	db_parse	ARGS((XPtrV *argsp, XPtrV *varps));
-static void	db_oaexpr	ARGS((XPtrV *argsp, XPtrV *varps));
-static void	db_nexpr	ARGS((XPtrV *argsp, XPtrV *varps));
-static void	db_primary	ARGS((XPtrV *argsp, XPtrV *varps));
+static int	dbtestp_isa ARGS((Test_env *te, Test_meta meta));
+static char	*dbtestp_getopnd ARGS((Test_env *te, Test_op op, int do_eval));
+static int	dbtestp_eval ARGS((Test_env *te, Test_op op, char *opnd1,
+				char *opnd2, int do_eval));
+static void	dbtestp_error ARGS((Test_env *te, int offset, char *msg));
 #endif /* KSH */
 
 static	struct	op	*outtree; /* yyparse output */
@@ -64,7 +66,7 @@ yyparse()
 		int c;
 		outtree = c_list();
 		if ((c = token(0)) != '\n' && c != 0)
-			syntaxerr();
+			syntaxerr((char *) 0);
 	}
 }
 
@@ -78,7 +80,7 @@ pipeline(cf)
 	if (t != NULL) {
 		while (token(0) == '|') {
 			if ((p = get_command(CONTIN)) == NULL)
-				syntaxerr();
+				syntaxerr((char *) 0);
 			if (tl == NULL)
 				t = tl = block(TPIPE, t, p, NOWORDS);
 			else
@@ -99,7 +101,7 @@ andor()
 	if (t != NULL) {
 		while ((c = token(0)) == LOGAND || c == LOGOR) {
 			if ((p = pipeline(CONTIN)) == NULL)
-				syntaxerr();
+				syntaxerr((char *) 0);
 			t = block(c == LOGAND? TAND: TOR, t, p, NOWORDS);
 		}
 		REJECT;
@@ -169,7 +171,7 @@ musthave(c, cf)
 	int c, cf;
 {
 	if ((token(cf)) != c)
-		syntaxerr();
+		syntaxerr((char *) 0);
 }
 
 static struct op *
@@ -232,7 +234,7 @@ get_command(cf)
 				if (iopn == 0 && XPsize(vars) == 0
 				    && XPsize(args) == 0
 				    && assign_command(ident))
-					t->evalflags = DOASNTILDE;
+					t->evalflags = DOVACHECK;
 				if ((XPsize(args) == 0 || Flag(FKEYWORD))
 				    && is_wdvarassign(yylval.cp))
 					XPput(vars, yylval.cp);
@@ -252,7 +254,7 @@ get_command(cf)
 				/* Must be a function */
 				if (iopn != 0 || XPsize(args) != 1
 				    || XPsize(vars) != 0)
-					syntaxerr();
+					syntaxerr((char *) 0);
 				ACCEPT;
 				/*(*/
 				musthave(')', 0);
@@ -293,7 +295,18 @@ get_command(cf)
 		syniocf &= ~(KEYWORD|ALIAS);
 		t = newtp(TDBRACKET);
 		ACCEPT;
-		db_parse(&args, &vars);
+		{
+			Test_env te;
+
+			te.flags = TEF_DBRACKET;
+			te.pos.av = &args;
+			te.isa = dbtestp_isa;
+			te.getopnd = dbtestp_getopnd;
+			te.eval = dbtestp_eval;
+			te.error = dbtestp_error;
+
+			test_parse(&te);
+		}
 		break;
 #endif /* KSH */
 
@@ -344,7 +357,7 @@ get_command(cf)
 		syniocf &= ~(KEYWORD|ALIAS);
 		t = pipeline(0);
 		if (t == (struct op *) 0)
-			syntaxerr();
+			syntaxerr((char *) 0);
 		t = block(TBANG, NOBLOCK, t, NOWORDS);
 		break;
 
@@ -406,7 +419,7 @@ dogroup()
 	else if (c == '{')
 		c = '}';
 	else
-		syntaxerr();
+		syntaxerr((char *) 0);
 	list = c_list();
 	musthave(c, KEYWORD|ALIAS);
 	return list;
@@ -421,7 +434,7 @@ thenpart()
 	t = newtp(0);
 	t->left = c_list();
 	if (t->left == NULL)
-		syntaxerr();
+		syntaxerr((char *) 0);
 	t->right = elsepart();
 	return (t);
 }
@@ -434,7 +447,7 @@ elsepart()
 	switch (token(KEYWORD|ALIAS|VARASN)) {
 	  case ELSE:
 		if ((t = c_list()) == NULL)
-			syntaxerr();
+			syntaxerr((char *) 0);
 		return (t);
 
 	  case ELIF:
@@ -565,7 +578,7 @@ wordlist()
 	while ((c = token(0)) == LWORD)
 		XPput(args, yylval.cp);
 	if (c != '\n' && c != ';')
-		syntaxerr();
+		syntaxerr((char *) 0);
 	if (XPsize(args) == 0) {
 		XPfree(args);
 		return NULL;
@@ -653,14 +666,16 @@ initkeywords()
 }
 
 static void
-syntaxerr()
+syntaxerr(what)
+	char *what;
 {
 	char redir[6];	/* 2<<- is the longest redirection, I think */
 	char *s;
-	char *what = "unexpected";
 	struct tokeninfo const *tt;
 	int c;
 
+	if (!what)
+		what = "unexpected";
 	REJECT;
 	c = token(0);
     Again:
@@ -775,132 +790,117 @@ assign_command(s)
 
 
 #ifdef KSH
-# define db_makevar(c)	(db_fakearg[1] = (c), wdcopy(db_fakearg, ATEMP))
-
-/* used by db_makevar() */
-static char db_fakearg[3] = { CHAR, 'x', EOS };
-
-/*
- * Parse a [[ ]] expression, converting it to a [ .. ] style expression,
- * saving the result in t->args.
+/* Order important - indexed by Test_meta values
+ * Note that ||, &&, ( and ) can't appear in as unquoted strings
+ * in normal shell input, so these can be interpreted unambiguously
+ * in the evaluation pass.
  */
-static void
-db_parse(argsp, varsp)
-	XPtrV *argsp, *varsp;
+static const char dbtest_or[] = { CHAR, '|', CHAR, '|', EOS };
+static const char dbtest_and[] = { CHAR, '&', CHAR, '&', EOS };
+static const char dbtest_not[] = { CHAR, '!', EOS };
+static const char dbtest_oparen[] = { CHAR, '(', EOS };
+static const char dbtest_cparen[] = { CHAR, ')', EOS };
+const char *const dbtest_tokens[] = {
+			dbtest_or, dbtest_and, dbtest_not,
+			dbtest_oparen, dbtest_cparen
+		};
+const char db_close[] = { CHAR, ']', CHAR, ']', EOS };
+const char db_lthan[] = { CHAR, '<', EOS };
+const char db_gthan[] = { CHAR, '>', EOS };
+
+/* Test if the current token is a whatever.  Accepts the current token if
+ * it is.  Returns 0 if it is not, non-zero if it is (in the case of
+ * TM_UNOP and TM_BINOP, the returned value is a Test_op).
+ */
+static int
+dbtestp_isa(te, meta)
+	Test_env *te;
+	Test_meta meta;
 {
-	static const char db_start[] = { CHAR, '[', CHAR, '[', EOS };
-	static const char db_close[] = { CHAR, ']', CHAR, ']', EOS };
+	int c = tpeek(ARRAYVAR | (meta == TM_BINOP ? 0 : CONTIN));
+	int uqword = 0;
+	char *save = (char *) 0;
+	int ret = 0;
 
-	XPput(*argsp, wdcopy(db_start, ATEMP));
-	XPput(*varsp, db_makevar(DB_NORM));
-	db_oaexpr(argsp, varsp);
-	if (tpeek(ARRAYVAR|CONTIN) != LWORD || strcmp(yylval.cp, db_close) != 0)
-		syntaxerr();
-	ACCEPT;
-}
+	/* unquoted word? */
+	uqword = c == LWORD && *ident;
 
-static void
-db_oaexpr(argsp, varsp)
-	XPtrV *argsp, *varsp;
-{
-	static const char or[] = { CHAR, '-', CHAR, 'o', EOS };
-	static const char and[] = { CHAR, '-', CHAR, 'a', EOS };
-	int c;
-
-	db_nexpr(argsp, varsp);
-	if ((c = tpeek(ARRAYVAR|CONTIN)) == LOGOR || c == LOGAND) {
+	if (meta == TM_OR)
+		ret = c == LOGOR;
+	else if (meta == TM_AND)
+		ret = c == LOGAND;
+	else if (meta == TM_NOT)
+		ret = uqword && strcmp(yylval.cp, dbtest_tokens[(int) TM_NOT]) == 0;
+	else if (meta == TM_OPAREN)
+		ret = c == '(' /*)*/;
+	else if (meta == TM_CPAREN)
+		ret = c == /*(*/ ')';
+	else if (meta == TM_UNOP || meta == TM_BINOP) {
+		if (meta == TM_BINOP && c == REDIR
+		    && (yylval.iop->flag == IOREAD
+			|| yylval.iop->flag == IOWRITE))
+		{
+			ret = 1;
+			save = wdcopy(yylval.iop->flag == IOREAD ?
+				db_lthan : db_gthan, ATEMP);
+		} else if (uqword && (ret = (int) test_isop(te, meta, ident)))
+			save = yylval.cp;
+	} else /* meta == TM_END */
+		ret = uqword && strcmp(yylval.cp, db_close) == 0;
+	if (ret) {
 		ACCEPT;
-		XPput(*argsp, wdcopy(c == LOGOR ? or : and, ATEMP));
-		XPput(*varsp, db_makevar(c == LOGOR ? DB_OR : DB_AND));
-		db_oaexpr(argsp, varsp);
-	}
-}
-
-static void
-db_nexpr(argsp, varsp)
-	XPtrV *argsp, *varsp;
-{
-	static const char not[] = { CHAR, '!', EOS };
-
-	if (tpeek(ARRAYVAR|CONTIN) == LWORD && strcmp(yylval.cp, not) == 0) {
-		ACCEPT;
-		XPput(*argsp, yylval.cp);
-		XPput(*varsp, db_makevar(DB_NORM));
-		db_nexpr(argsp, varsp);
-	} else
-		db_primary(argsp, varsp);
-}
-
-static void
-db_primary(argsp, varsp)
-	XPtrV *argsp, *varsp;
-{
-	static const char oparen[] = { CHAR, '(', EOS };
-	static const char cparen[] = { CHAR, ')', EOS };
-	int c;
-
-	c = token(ARRAYVAR|CONTIN);
-	if (c == '(' /*)*/) {
-		XPput(*argsp, wdcopy(oparen, ATEMP));
-		XPput(*varsp, db_makevar(DB_NORM));
-		db_oaexpr(argsp, varsp);
-		/*(*/
-		if (token(ARRAYVAR|CONTIN) != ')')
-			/*(*/
-			yyerror("missing )\n");
-		XPput(*argsp, wdcopy(cparen, ATEMP));
-		XPput(*varsp, db_makevar(DB_NORM));
-	} else if (c == LWORD) {
-		if (*ident && is_db_unop(ident)) {
-			XPput(*argsp, yylval.cp);
-			XPput(*varsp, db_makevar(DB_NORM));
-			if (token(ARRAYVAR) != LWORD)
-				syntaxerr();
-			XPput(*argsp, yylval.cp);
-			XPput(*varsp, db_makevar(DB_NORM));
-		} else {
-			static const char binop_marker[] =
-				    { CHAR, '-', CHAR, 'B', CHAR, 'E', EOS };
-
-			/* must be a binary operator: mark this
-			 * with special -BE operator so we can't confuse
-			 * the binary expression '-z = foobar' for
-			 * something else. (-BE is only recognized
-			 * when parsing [[ ]] expressions)
-			 */
-			XPput(*argsp, wdcopy(binop_marker, ATEMP));
-			XPput(*varsp, db_makevar(DB_BE));
-			XPput(*argsp, yylval.cp);
-			XPput(*varsp, db_makevar(DB_NORM));
-			c = token(ARRAYVAR);
-			/* convert REDIR < and > to LWORD < and > */
-			if (c == REDIR
-			    && (yylval.iop->flag == IOREAD
-				|| yylval.iop->flag == IOWRITE))
-			{
-				static const char lthan[] = { CHAR, '<', EOS };
-				static const char gthan[] = { CHAR, '>', EOS };
-				const char *what = yylval.iop->flag == IOREAD ?
-						lthan : gthan;
-
-				afree(yylval.iop, ATEMP);
-				yylval.cp = wdcopy(what, ATEMP);
-				c = symbol = LWORD;
-				ident[0] = what[1];
-				ident[1] = '\0';
-			}
-			if (c == LWORD && *ident && (c = is_db_binop(ident))) {
-				XPput(*argsp, yylval.cp);
-				XPput(*varsp, db_makevar(DB_NORM));
-				if (token(ARRAYVAR) != LWORD)
-					syntaxerr();
-				XPput(*argsp, yylval.cp);
-				XPput(*varsp, db_makevar(is_db_patop(c) ?
-							    DB_PAT : DB_NORM));
-			} else
-				syntaxerr();
+		if (meta != TM_END) {
+			if (!save)
+				save = wdcopy(dbtest_tokens[(int) meta], ATEMP);
+			XPput(*te->pos.av, save);
 		}
-	} else
-		syntaxerr();
+	}
+	return ret;
+}
+
+static char *
+dbtestp_getopnd(te, op, do_eval)
+	Test_env *te;
+	Test_op op;
+	int do_eval;
+{
+	int c = tpeek(ARRAYVAR);
+
+	if (c != LWORD)
+		return (char *) 0;
+
+	ACCEPT;
+	XPput(*te->pos.av, yylval.cp);
+
+	return null;
+}
+
+static int
+dbtestp_eval(te, op, opnd1, opnd2, do_eval)
+	Test_env *te;
+	Test_op op;
+	char *opnd1;
+	char *opnd2;
+	int do_eval;
+{
+	return 1;
+}
+
+static void
+dbtestp_error(te, offset, msg)
+	Test_env *te;
+	int offset;
+	char *msg;
+{
+	te->flags |= TEF_ERROR;
+
+	if (offset < 0) {
+		REJECT;
+		/* Kludgy to say the least... */
+		symbol = LWORD;
+		yylval.cp = *(XPptrv(*te->pos.av) + XPsize(*te->pos.av)
+				+ offset);
+	}
+	syntaxerr(msg);
 }
 #endif /* KSH */
