@@ -72,23 +72,10 @@ ptree(t, indent, shf)
 	  {
 		int i;
 
-		/* the opening [[ is in args[0] */
+		fptreef(shf, indent, "[[");
 		for (i = 0; t->args[i]; i++)
-			switch (t->vars[i][1]) {
-			  case DB_NORM:
-			  case DB_PAT:
-				fptreef(shf, indent, "%S ", t->args[i]);
-				break;
-			  case DB_OR:
-				fptreef(shf, indent, "|| ");
-				break;
-			  case DB_AND:
-				fptreef(shf, indent, "&& ");
-			  case DB_BE:
-				/* print nothing */
-				break;
-			}
-		fptreef(shf, indent, "]] ");
+			fptreef(shf, indent, " %S", t->args[i]);
+		fptreef(shf, indent, " ]] ");
 		break;
 	  }
 	  case TFOR:
