@@ -35,8 +35,13 @@ newblock()
 
 	l = (struct block *) alloc(sizeof(struct block), ATEMP);
 	ainit(&l->area);
-	l->argc = 0;
-	l->argv = empty;
+	if (!e->loc) {
+		l->argc = 0;
+		l->argv = empty;
+	} else {
+		l->argc = e->loc->argc;
+		l->argv = e->loc->argv;
+	}
 	l->exit = l->error = NULL;
 	tinit(&l->vars, &l->area, 0);
 	tinit(&l->funs, &l->area, 0);
@@ -55,7 +60,6 @@ popblock()
 	register int i;
 
 	e->loc = l->next;	/* pop block */
-	kshname = e->loc->argv[0];
 	for (i = l->vars.size; --i >= 0; )
 		if ((vp = *vpp++) != NULL && (vp->flag&SPECIAL))
 			if ((vq = global(vp->name))->flag & ISSET)
@@ -123,6 +127,7 @@ global(n)
 	p = skip_varname(n, FALSE);
 	if (p != n && *p == '[' && (len = array_ref_len(p))) {
 		char *sub, *tmp;
+		long rval;
 
 		/* Calculate the value of the subscript */
 		array = 1;
@@ -130,7 +135,8 @@ global(n)
 		sub = substitute(tmp, 0);
 		afree(tmp, ATEMP);
 		n = strnsave(n, p - n, ATEMP);
-		val = evaluate(sub);
+		evaluate(sub, &rval, FALSE);
+		val = rval;
 		if (val < 0 || val > ARRAYMAX)
 			errorf("%s: subscript out of range", n);
 		afree(sub, ATEMP);
@@ -222,6 +228,7 @@ local(n)
 	p = skip_varname(n, FALSE);
 	if (p != n && *p == '[' && (len = array_ref_len(p))) {
 		char *sub, *tmp;
+		long rval;
 
 		/* Calculate the value of the subscript */
 		array = 1;
@@ -229,7 +236,8 @@ local(n)
 		sub = substitute(tmp, 0);
 		afree(tmp, ATEMP);
 		n = strnsave(n, p - n, ATEMP);
-		val = evaluate(sub);
+		evaluate(sub, &rval, FALSE);
+		val = rval;
 		if (val < 0 || val > ARRAYMAX)
 			errorf("%s: subscript out of range", n);
 		afree(sub, ATEMP);
@@ -342,7 +350,8 @@ setstr(vq, s)
 				vq->flag |= ALLOC;
 		}
 	} else			/* integer dest */
-		v_evaluate(vq, s);
+		/* XXX is this correct? */
+		v_evaluate(vq, s, FALSE);
 	vq->flag |= ISSET;
 	if ((vq->flag&SPECIAL))
 		setspec(vq);
@@ -587,10 +596,8 @@ typeset(var, set, clr, field, base)
 
 	vp = (set&LOCAL) ? local(tvar) : global(tvar);
 	set &= ~LOCAL;
-	if (val)
-		afree(tvar, ATEMP);
 
-	vpbase = (vp->flag & ARRAY) ? global(basename(var)) : vp;
+	vpbase = (vp->flag & ARRAY) ? global(arrayname(var)) : vp;
 
 	/* only allow export flag to be set.  at&t ksh allows any attribute to
 	 * be changed, which means it can be truncated or modified
@@ -599,7 +606,9 @@ typeset(var, set, clr, field, base)
 	if ((vpbase->flag&RDONLY)
 	    && (val || clr || (set & ~EXPORT)))
 		/* XXX check calls - is error here ok by POSIX? */
-		errorf("%s: is read only", var);
+		errorf("%s: is read only", tvar);
+	if (val)
+		afree(tvar, ATEMP);
 
 	/* most calls are with set/clr == 0 */
 	if (set | clr) {
@@ -665,12 +674,29 @@ typeset(var, set, clr, field, base)
 	return vp;
 }
 
+/* Unset a variable.  array_ref is set if there was an array reference in
+ * the name lookup (eg, x[2]).
+ */
 void
-unset(vp)
+unset(vp, array_ref)
 	register struct tbl *vp;
+	int array_ref;
 {
-	if ((vp->flag&ALLOC))
+	if (vp->flag & ALLOC)
 		afree((void*)vp->val.s, vp->areap);
+	if ((vp->flag & ARRAY) && !array_ref) {
+		struct tbl *a, *tmp;
+
+		/* Free up entire array */
+		for (a = vp->u.array; a; ) {
+			tmp = a;
+			a = a->u.array;
+			if (tmp->flag & ALLOC)
+				afree((void *) tmp->val.s, tmp->areap);
+			afree(tmp, tmp->areap);
+		}
+		vp->u.array = (struct tbl *) 0;
+	}
 	vp->flag &= SPECIAL;	/* Should ``unspecial'' some vars */
 	if (vp->flag & SPECIAL)
 		unsetspec(vp);
@@ -875,7 +901,7 @@ setspec(vp)
 		{
 			struct stat statb;
 			s = strval(vp);
-			if (ISDIRSEP(s[0]) && eaccess(s, W_OK|X_OK) == 0
+			if (ISABSPATH(s) && eaccess(s, W_OK|X_OK) == 0
 			    && stat(s, &statb) == 0 && S_ISDIR(statb.st_mode))
 				tmpdir = strsave(s, APERM);
 		}
@@ -908,8 +934,8 @@ setspec(vp)
 			set_editmode(strval(vp));
 		break;
 	  case V_COLUMNS:
-		if ((x_cols = intval(vp)) <= 0)
-			x_cols = 16;
+		if ((x_cols = intval(vp)) <= MIN_COLS)
+			x_cols = MIN_COLS;
 		break;
 # endif /* EDIT */
 	  case V_TMOUT:
@@ -1039,7 +1065,7 @@ array_ref_len(cp)
  * Make a copy of the base of an array name
  */
 char *
-basename(str)
+arrayname(str)
 	char	*str;
 {
 	char	*p;
@@ -1071,7 +1097,7 @@ set_array(var, reset, vals)
 	/* This code is quite non-optimal */
 	if (reset > 0)
 		/* trash existing values and attributes */
-		unset(vp);
+		unset(vp, 0);
 	for (i = 0; vals[i]; i++) {
 		vq = arraysearch(vp, i);
 		setstr(vq, vals[i]);
