@@ -16,8 +16,10 @@
 #define Ctrl(c)		(c&0x1f)
 #define	iswordch(c)	(letnum(c))
 
+
 typedef struct glob Glob;
 
+static int	vi_hook	ARGS((int ch));
 static void 	vi_reset ARGS((char *buf, size_t len));
 static int	nextstate ARGS((int ch));
 static int	vi_insert ARGS((int ch));
@@ -54,14 +56,16 @@ static int	complete_word ARGS((int command));
 static int	print_expansions ARGS((int command));
 static int 	char_len ARGS((int c));
 static void 	x_vi_zotc ARGS((int c));
+static void	vi_error ARGS((void));
+static void	vi_macro_reset ARGS((void));
 
-#define C_	0x1
-#define M_	0x2
-#define E_	0x4
-#define X_	0x8
-#define U_	0x10
-#define B_	0x20
-#define S_	0x80
+#define C_	0x1		/* a valid command that isn't a M_, E_, U_ */
+#define M_	0x2		/* movement command (h, l, etc.) */
+#define E_	0x4		/* extended command (c, d, y) */
+#define X_	0x8		/* long command (@, f, F, t, T, etc.) */
+#define U_	0x10		/* an UN-undoable command (that isn't a M_) */
+#define B_	0x20		/* bad command (^@) */
+#define S_	0x80		/* search (/, ?) */
 
 #define isbad(c)	(classify[(c)&0x7f]&B_)
 #define iscmd(c)	(classify[(c)&0x7f]&(M_|E_|C_|U_))
@@ -72,23 +76,39 @@ static void 	x_vi_zotc ARGS((int c));
 #define issrch(c)	(classify[(c)&0x7f]&S_)
 
 char	classify[128] = {
-	     /*	0	1	2	3	4	5	6	7  */
-   /*   0 */	B_,	0,	0,	0,	0,	C_|U_,	C_|U_,	0,
-   /*  01 */	C_|M_,	0,	0,	0,	C_|U_,	0,	C_,	0,
-   /*  02 */	C_,	0,	C_|U_,	0,	0,	0,	0,	0,
-   /*  03 */	0,	0,	C_,	0,	0,	0,	0,	0,
-   /*  04 */	M_,	0,	0,	C_,	M_,	M_,	0,	0,
-   /*  05 */	0,	0,	C_,	C_,	M_,	C_,	0,	C_|S_,
-   /*  06 */	M_,	0,	0,	0,	0,	0,	0,	0,
-   /*  07 */	0,	0,	0,	M_,	0,	C_,	0,	C_|S_,
-   /* 010 */	0,	C_,	M_,	C_,	C_,	M_,	M_|X_,	C_,
-   /* 011 */	0,	C_,	0,	0,	0,	0,	C_,	0,
-   /* 012 */	C_,	0,	C_,	C_,	M_|X_,	0,	0,	M_,
-   /* 013 */	C_,	C_,	0,	0,	C_,	0,	M_,	C_,
-   /* 014 */	0,	C_,	M_,	E_,	E_,	M_,	M_|X_,	C_,
-   /* 015 */	M_,	C_,	C_,	C_,	M_,	0,	C_,	0,
-   /* 016 */	C_,	0,	X_,	C_,	M_|X_,	C_|U_,	C_,	M_,
-   /* 017 */	C_,	E_,	0,	0,	M_,	0,	C_,	0
+   /*           0       1       2       3       4       5       6       7    */
+   /*   0       ^@     ^A      ^B      ^C      ^D      ^E      ^F      ^G    */
+		B_,     0,      0,      0,      0,      C_|U_,  C_,     0,
+   /*  01       ^H     ^I      ^J      ^K      ^L      ^M      ^N      ^O    */
+		M_,     0,      0,      0,      C_|U_,  0,      C_,     0,
+   /*  02       ^P     ^Q      ^R      ^S      ^T      ^U      ^V      ^W    */
+		C_,     0,      C_|U_,  0,      0,      0,      C_,     0,
+   /*  03       ^X     ^Y      ^Z      ^[      ^\      ^]      ^^      ^_    */
+		C_,     0,      0,      0,      0,      0,      0,      0,
+   /*  04      <space>  !       "       #       $       %       &       '    */
+		M_,     0,      0,      C_,     M_,     M_,     0,      0,
+   /*  05       (       )       *       +       ,       -       .       /    */
+		0,      0,      C_,     C_,     M_,     C_,     0,      C_|S_,
+   /*  06       0       1       2       3       4       5       6       7    */
+		M_,     0,      0,      0,      0,      0,      0,      0,
+   /*  07       8       9       :       ;       <       =       >       ?    */
+		0,      0,      0,      M_,     0,      C_,     0,      C_|S_,
+   /* 010       @       A       B       C       D       E       F       G    */
+		C_|X_,  C_,     M_,     C_,     C_,     M_,     M_|X_,  C_|U_,
+   /* 011       H       I       J       K       L       M       N       O    */
+		0,      C_,     0,      0,      0,      0,      C_|U_,  0,
+   /* 012       P       Q       R       S       T       U       V       W    */
+		C_,     0,      C_,     C_,     M_|X_,  0,      0,      M_,
+   /* 013       X       Y       Z       [       \       ]       ^       _    */
+		C_,     C_|U_,  0,      0,      C_,     0,      M_,     C_,
+   /* 014       `       a       b       c       d       e       f       g    */
+		0,      C_,     M_,     E_,     E_,     M_,     M_|X_,  C_,
+   /* 015       h       i       j       k       l       m       n       o    */
+		M_,     C_,     C_|U_,  C_|U_,  M_,     0,      C_|U_,  0,
+   /* 016       p       q       r       s       t       u       v       w    */
+		C_,     0,      X_,     C_,     M_|X_,  C_|U_,  C_|U_,  M_,
+   /* 017       x       y       z       {       |       }       ~      ^?    */
+		C_,     E_|U_,  0,      0,      M_,     0,      C_,     0
 };
 
 #define MAXVICMD	3
@@ -101,12 +121,13 @@ char	classify[128] = {
 #define VARG1		1		/* digit prefix (first, eg, 5l) */
 #define VEXTCMD		2		/* cmd + movement (eg, cl) */
 #define VARG2		3		/* digit prefix (second, eg, 2c3l) */
-#define VXCH		4		/* f, F, t, T */
+#define VXCH		4		/* f, F, t, T, @ */
 #define VFAIL		5		/* bad command */
 #define VCMD		6		/* single char command (eg, X) */
 #define VREDO		7		/* . */
 #define VLIT		8		/* ^V */
 #define VSEARCH		9		/* /, ? */
+#define VVERSION	10		/* <ESC> ^V */
 
 struct edstate {
 	int	winleft;
@@ -135,6 +156,8 @@ static struct edstate	*es;			/* current editor state */
 static struct edstate	*undo;
 
 static char	ibuf[CMDLEN];		/* input buffer */
+static int	first_insert;		/* set when starting in insert mode */
+static int	saved_inslen;		/* saved inslen for first insert */
 static int	inslen;			/* length of input buffer */
 static int	srchlen;		/* length of current search pattern */
 static char	ybuf[CMDLEN];		/* yank buffer */
@@ -151,10 +174,23 @@ static int	hlast;			/* 1 past last position in history */
 static int	modified;		/* buffer has been "modified" */
 static int	state;
 
+/* Information for keeping track of macros that are being expanded.
+ * The format of buf is the alias contents followed by a null byte followed
+ * by the name (letter) of the alias.  The end of the buffer is marked by
+ * a double null.  The name of the alias is stored so recursive macros can
+ * be detected.
+ */
+struct macro_state {
+    unsigned char	*p;	/* current position in buf */
+    unsigned char	*buf;	/* pointer to macro(s) being expanded */
+    int			len;	/* how much data in buffer */
+};
+static struct macro_state macro;
+
 enum expand_mode { NONE, EXPAND, COMPLETE, PRINT };
 static enum expand_mode expanded = NONE;/* last input was expanded */
 
-int
+static int
 vi_hook(ch)
 	int		ch;
 {
@@ -173,8 +209,32 @@ vi_hook(ch)
 			}
 			switch (vi_insert(ch)) {
 			case -1:
-				x_putc(BEL);
+#ifdef OS2
+				/* Arrow keys generate 0xe0X, where X is H.. */
+				state = VCMD;
+				argc1 = 1;
+				switch (x_getc()) {
+				  case 'H':
+					*curcmd='k';
+					break;
+				  case 'K':
+					*curcmd='h';
+					break;
+				  case 'P':
+					*curcmd='j';
+					break;
+				  case 'M':
+					*curcmd='l';
+					break;
+				  default:
+					vi_error();
+					state = VNORMAL;
+				}
+				break;
+#else /* OS2 */
+				vi_error();
 				state = VNORMAL;
+#endif /* OS2 */
 				break;
 			case 0:
 				if (state == VLIT) {
@@ -205,10 +265,18 @@ vi_hook(ch)
 						if (putbuf("/", 1, 0) != 0) {
 							return -1;
 						}
-					} else if (putbuf("?", 1, 0) != 0) 
+					} else if (putbuf("?", 1, 0) != 0)
 							return -1;
 					refresh(0);
 				}
+				if (state == VVERSION) {
+					save_cbuf();
+					es->cursor = 0;
+					es->linelen = 0;
+					putbuf(ksh_version + 4,
+						strlen(ksh_version + 4), 0);
+					refresh(0);
+				}
 			}
 		}
 		break;
@@ -216,13 +284,19 @@ vi_hook(ch)
 	case VLIT:
 		if (isbad(ch)) {
 			del_range(es->cursor, es->cursor + 1);
-			x_putc(BEL);
+			vi_error();
 		} else
 			es->cbuf[es->cursor++] = ch;
 		refresh(1);
 		state = VNORMAL;
 		break;
 
+	case VVERSION:
+		restore_cbuf();
+		state = VNORMAL;
+		refresh(0);
+		break;
+
 	case VARG1:
 		if (isdigit(ch))
 			argc1 = argc1 * 10 + ch - '0';
@@ -315,7 +389,7 @@ vi_hook(ch)
 			return 0;
 		} else {
 			if (srchlen == SRCHLEN - 1)
-				x_putc(BEL);
+				vi_error();
 			else {
 				locpat[srchlen++] = ch;
 				if ((ch & 0x80) && Flag(FVISHOW8)) {
@@ -341,7 +415,7 @@ vi_hook(ch)
 		state = VNORMAL;
 		switch (vi_cmd(argc1, curcmd)) {
 		case -1:
-			x_putc(BEL);
+			vi_error();
 			break;
 		case 0:
 			if (insert != 0)
@@ -363,7 +437,7 @@ vi_hook(ch)
 			lastac = argc1;
 		switch (vi_cmd(lastac, lastcmd) != 0) {
 		case -1:
-			x_putc(BEL);
+			vi_error();
 			refresh(0);
 			break;
 		case 0:
@@ -371,10 +445,10 @@ vi_hook(ch)
 				if (lastcmd[0] == 's' || lastcmd[0] == 'c' ||
 						lastcmd[0] == 'C') {
 					if (redo_insert(1) != 0)
-						x_putc(BEL);
+						vi_error();
 				} else {
 					if (redo_insert(lastac) != 0)
-						x_putc(BEL);
+						vi_error();
 				}
 			}
 			refresh(0);
@@ -390,7 +464,7 @@ vi_hook(ch)
 
 	case VFAIL:
 		state = VNORMAL;
-		x_putc(BEL);
+		vi_error();
 		break;
 	}
 	return 0;
@@ -404,11 +478,11 @@ vi_reset(buf, len)
 	state = VNORMAL;
 	hnum = hlast = histnum(-1) + 1;
 	insert = INSERT;
-	yanklen = 0;
+	saved_inslen = inslen;
+	first_insert = 1;
 	inslen = 0;
-	lastcmd[0] = 'a';
-	lastac = 1;
 	modified = 1;
+	vi_macro_reset();
 	edit_reset(buf, len);
 }
 
@@ -424,6 +498,8 @@ nextstate(ch)
 		return VXCH;
 	else if (ch == '.')
 		return VREDO;
+	else if (ch == Ctrl('v'))
+		return VVERSION;
 	else if (iscmd(ch))
 		return VCMD;
 	else
@@ -439,7 +515,7 @@ vi_insert(ch)
 	if (ch == edchars.erase || ch == Ctrl('h')) {
 		if (insert == REPLACE) {
 			if (es->cursor == undo->cursor) {
-				x_putc(BEL);
+				vi_error();
 				return 0;
 			}
 			if (inslen > 0)
@@ -490,8 +566,17 @@ vi_insert(ch)
 		expanded = NONE;
 		return 0;
 	}
+	/* If any chars are entered before escape, trash the saved insert
+	 * buffer (if user inserts & deletes char, ibuf gets trashed and
+	 * we don't want to use it)
+	 */
+	if (first_insert && ch != Ctrl('['))
+		saved_inslen = 0;
 	switch (ch) {
 
+#ifdef OS2
+	case 224:	 /* function key prefix */
+#endif /* OS2 */
 	case '\0':
 		return -1;
 
@@ -501,6 +586,15 @@ vi_insert(ch)
 
 	case Ctrl('['):
 		expanded = NONE;
+		if (first_insert) {
+			first_insert = 0;
+			if (inslen == 0) {
+				inslen = saved_inslen;
+				return redo_insert(0);
+			}
+			lastcmd[0] = 'a';
+			lastac = 1;
+		}
 		if (lastcmd[0] == 's' || lastcmd[0] == 'c' ||
 				lastcmd[0] == 'C')
 			return redo_insert(0);
@@ -568,7 +662,8 @@ vi_cmd(argcnt, cmd)
 		} else
 			return -1;
 	} else {
-		if (isundoable(*cmd)) {
+		/* Don't save state in middle of macro.. */
+		if (isundoable(*cmd) && !macro.p) {
 			undo->winleft = es->winleft;
 			memmove(undo->cbuf, es->cbuf, es->linelen);
 			undo->linelen = es->linelen;
@@ -583,6 +678,43 @@ vi_cmd(argcnt, cmd)
 			redraw_line();
 			break;
 
+		case '@':
+			{
+				static char alias[] = "_\0";
+				struct tbl *ap;
+				int	olen, nlen;
+				char	*p, *nbuf;
+
+				/* lookup letter in alias list... */
+				alias[1] = cmd[1];
+				ap = tsearch(&aliases, alias, hash(alias));
+				if (!cmd[1] || !ap || !(ap->flag & ISSET))
+					return -1;
+				/* check if this is a recursive call... */
+				if ((p = (char *) macro.p))
+					while ((p = strchr(p, '\0')) && p[1])
+						if (*++p == cmd[1])
+							return -1;
+				/* insert alias into macro buffer */
+				nlen = strlen(ap->val.s) + 1;
+				olen = !macro.p ? 2
+					: macro.len - (macro.p - macro.buf);
+				nbuf = alloc(nlen + 1 + olen, APERM);
+				memcpy(nbuf, ap->val.s, nlen);
+				nbuf[nlen++] = cmd[1];
+				if (macro.p) {
+					memcpy(nbuf + nlen, macro.p, olen);
+					afree(macro.buf, APERM);
+					nlen += olen;
+				} else {
+					nbuf[nlen++] = '\0';
+					nbuf[nlen++] = '\0';
+				}
+				macro.p = macro.buf = (unsigned char *) nbuf;
+				macro.len = nlen;
+			}
+			break;
+
 		case 'a':
 			modified = 1; hnum = hlast;
 			if (es->linelen != 0)
@@ -631,6 +763,8 @@ vi_cmd(argcnt, cmd)
 				} else {
 					c1 = ncursor;
 					c2 = es->cursor;
+					if (cmd[1] == '%')
+						c2++;
 				}
 			}
 			if (*cmd != 'c' && c1 != c2)
@@ -735,7 +869,7 @@ vi_cmd(argcnt, cmd)
 				return -1;
 			modified = 1; hnum = hlast;
 			if (cmd[1] == 0)
-				x_putc(BEL);
+				vi_error();
 			else
 				es->cbuf[es->cursor] = cmd[1];
 			break;
@@ -816,7 +950,7 @@ vi_cmd(argcnt, cmd)
 			if (lastsearch == ' ')
 				return -1;
 			if (lastsearch == '?')
-				c1 = 1; 
+				c1 = 1;
 			else
 				c1 = 0;
 			if (*cmd == 'N')
@@ -919,7 +1053,7 @@ vi_cmd(argcnt, cmd)
 			print_expansions(1);
 			break;
 
-		/* case Ctrl('['):	some annoying at&t ksh's */
+		case Ctrl('['):			/* some annoying at&t ksh's */
 		case '\\':			/* at&t ksh */
 		case Ctrl('f'):			/* Nonstandard vi/ksh */
 			complete_word(1);
@@ -1077,7 +1211,7 @@ domove(argcnt, cmd, sub)
 			else if (t == -i)
 				bcount--;
 		} while (bcount != 0);
-		if (sub)
+		if (sub && i > 0)
 			ncursor++;
 		break;
 
@@ -1144,6 +1278,7 @@ bracktype(ch)
 
 static int	cur_col;		/* current column on line */
 static int	pwidth;			/* width of prompt */
+static int	prompt_trunc;		/* how much of prompt to truncate */
 static int	winwidth;		/* width of window */
 static char	*wbuf[2];		/* window buffers */
 static int	wbuf_len;		/* length of window buffers (x_cols-3)*/
@@ -1204,8 +1339,8 @@ free_edstate(old)
 	afree(old->cbuf, APERM);
 	afree((char *)old, APERM);
 }
-	
-	
+
+
 
 static void
 edit_reset(buf, len)
@@ -1223,6 +1358,12 @@ edit_reset(buf, len)
 	es->winleft = undo->winleft = 0;
 
 	cur_col = pwidth = promptlen(prompt);
+	if (pwidth > x_cols - 3 - MIN_EDIT_SPACE) {
+		cur_col = x_cols - 3 - MIN_EDIT_SPACE;
+		prompt_trunc = pwidth - cur_col;
+		pwidth -= prompt_trunc;
+	} else
+		prompt_trunc = 0;
 	if (!wbuf_len || wbuf_len != x_cols - 3) {
 		wbuf_len = x_cols - 3;
 		wbuf[0] = aresize(wbuf[0], wbuf_len, APERM);
@@ -1477,6 +1618,7 @@ grabsearch(save, start, fwd, pat)
 	anchored = *pat == '^' ? (++pat, 1) : 0;
 	if ((hist = findhist(start, fwd, pat, anchored)) < 0) {
 		/* if (start != 0 && fwd && match(holdbuf, pat) >= 0) { */
+		/* XXX should FILECMP be strncmp? */
 		if (start != 0 && fwd && FILECMP(holdbuf, pat) >= 0) {
 			restore_cbuf();
 			return 0;
@@ -1500,7 +1642,7 @@ redraw_line()
 	(void) memset(wbuf[win], ' ', wbuf_len);
 	x_putc('\r');
 	x_putc('\n');
-	pprompt(prompt);
+	pprompt(prompt, prompt_trunc);
 	cur_col = pwidth;
 	morec = ' ';
 }
@@ -1643,6 +1785,9 @@ display(wb1, wb2, leftside)
 		col++;
 	}
 	if (es->winleft > 0 && moreright)
+		/* POSIX says to use * for this but that is a globing
+		 * character and may confuse people; + is more innocuous
+		 */
 		mc = '+';
 	else if (es->winleft > 0)
 		mc = '<';
@@ -1668,7 +1813,7 @@ ed_mov_opt(col, wb)
 	if (col < cur_col) {
 		if (col + 1 < cur_col - col) {
 			x_putc('\r');
-			pprompt(prompt);
+			pprompt(prompt, prompt_trunc);
 			cur_col = pwidth;
 			while (cur_col++ < col)
 				x_putc(*wb++);
@@ -1690,11 +1835,25 @@ x_vi(buf, len)
 	size_t	len;
 {
 	int	c;
-	
+
 	vi_reset(buf, len > CMDLEN ? CMDLEN : len);
+	pprompt(prompt, prompt_trunc);
 	x_flush();
 	while (1) {
-		if ((c = x_getc()) == -1)
+		if (macro.p) {
+			c = *macro.p++;
+			/* end of current macro? */
+			if (!c) {
+				/* more macros left to finish? */
+				if (*macro.p++)
+					continue;
+				/* must be the end of all the macros */
+				vi_macro_reset();
+				c = x_getc();
+			}
+		} else
+			c = x_getc();
+		if (c == -1)
 			break;
 		if (state != VLIT) {
 			if (c == edchars.intr || c == edchars.quit) {
@@ -1704,7 +1863,7 @@ x_vi(buf, len)
 				trapsig(c == edchars.intr ? SIGINT : SIGQUIT);
 				x_mode(FALSE);
 				unwind(LSHELL);
-			} else if (c == edchars.eof) {
+			} else if (c == edchars.eof && state != VVERSION) {
 				if (es->linelen == 0) {
 					x_vi_zotc(edchars.eof);
 					c = -1;
@@ -1774,74 +1933,76 @@ glob_word(g, command)
 	int	command;
 {
 	int		rval;
-	int		star;
 	int		start, end;
-	char		*toglob = undo->cbuf;
+	char		*toglob;
+	char		*s;
 
 	if (!command && es->linelen > 0)
 		es->cursor--;
-	/*
-	 *	expansion normally maps a*b to `^a.*b.*'; if the cursor
-	 *  is just past the end of the pattern, then map a*b to `^a.*b$'.
-	 */
-	if (isspace(es->cbuf[es->cursor])) {
-		if (es->cursor == 0 || isspace(es->cbuf[es->cursor - 1])) {
-			if (!command && es->linelen > 0)
-				es->cursor++;
-			x_putc(BEL);
-			return -1;
-		}
-		start = es->cursor - 1;
-		star = 0;
-	} else {
-		start = es->cursor;
-		star = 1;
-	}
-	while (start > -1 && !isspace(es->cbuf[start]))
-		start--;
-	start++;
-	end = es->cursor;
-	while (end < es->linelen && !isspace(es->cbuf[end]))
-		end++;
-	if (start == end) {
+
+	start = es->cursor;
+	if (es->linelen == 0
+	    || (isspace(es->cbuf[es->cursor])
+		&& (--start < 0 || isspace(es->cbuf[start]))))
+	{
+		/* This is different from at&t ksh - it makes
+		 * a mountain out of a molehill (ie, '' -> *).
+		 * XXX what does posix say?
+		 */
 		if (!command && es->linelen > 0)
 			es->cursor++;
+		vi_error();
 		return -1;
 	}
 
+	while (start > 0 && !isspace(es->cbuf[start - 1]))
+		start--;
+	for (end = es->cursor; end < es->linelen && !isspace(es->cbuf[end]); )
+		end++;
+
 	/* use undo buffer to build word up in */
-	memmove(toglob, &es->cbuf[start], end-start);
-	if (star && toglob[end-start-1] != '*') {
-		toglob[end-start] = '*';
-		toglob[end-start+1] = '\0';
-	} else
-		toglob[end-start] = '\0';
+	memmove(toglob = undo->cbuf, &es->cbuf[start], end - start);
+	toglob[end - start] = '\0';
+
+	/*
+	 * Even though ESC-* is documented as globbing the current word
+	 * as if a was appended to it, that's not the way it works under
+	 * at&t ksh.  If the pathname contains a wildcard
+	 * (an unquoted '*', '?', or '['), then it is globbed based on
+	 * that value (i.e., without the appended '*').
+	 */
+	for (s = toglob; *s; s++) {
+		if (*s == '\\' && s[1])
+			s++;
+		else if (*s == '*' || *s == '[' || *s == '?')
+			break;
+	}
+	if (!*s) {
+		toglob[end - start] = '*';
+		toglob[end - start + 1] = '\0';
+	}
 
 	g->start = start;
 	g->end = end;
 	g->result = globstr(toglob);
-	g->addspace = 0;
-	rval = 0;
+	g->addspace = isspace(es->cbuf[es->cursor]);
 
+	rval = 0;
 	if (strcmp(g->result[0], toglob) == 0 && g->result[1] == (char *) 0) {
 		struct stat	statb;
 
-		/* If no star was appended, we need an additional check to
-		 * see if this really is an error.  E.g. toglob == "abc"
-		 * when the file "abc" exists is not an error; if
-		 * "abc" does exist it is an error.
+		/* check to see if this really is an error.
+		 * E.g. toglob == "ab*" when the file "ab*" exists
+		 * is not an error.
 		 */
-		if (star || stat(g->result[0], &statb) < 0) {
+		if (stat(g->result[0], &statb) < 0) {
 			if (!command && es->linelen > 0)
 				es->cursor++;
-			x_putc(BEL);
+			vi_error();
 			rval = -1;
 		}
 	}
 
-	if (rval == 0 && isspace(es->cbuf[es->cursor]))
-		g->addspace = 1;
-
 	/* restore undo buffer that we used temporarily */
 	memmove(toglob, es->cbuf, es->linelen);
 	return rval;
@@ -1899,7 +2060,7 @@ complete_word(command)
 {
 	int	rval = 0;
 	Glob	g;
-	int	i, escape = 0, len;
+	int	i, len;
 	struct	stat s;
 	static 	struct edstate *buf = 0;
 
@@ -1919,24 +2080,6 @@ complete_word(command)
 	}
 	if (glob_word(&g, command) < 0)
 		return -1;
-	/* Check for globbing in arg string */
-	for (i=g.start; i<g.end; i++) {
-		switch (es->cbuf[i]) {
-		case '\\':
-			escape = 1;
-			break;
-		case '*':
-		case '?':
-		case '[':
-			if (!escape) {
-				x_putc(BEL);
-				return -1;
-			}
-			break;
-		default:
-			escape = 0;
-		}
-	}
 
 	buf = save_edstate(es);
 	expanded = COMPLETE;
@@ -1946,9 +2089,11 @@ complete_word(command)
 		/* find the shortest match */
 		char	*save = g.result[0], *p;
 		len = strlen(g.result[0]);
-		if (g.result[1] != 0) 
-			x_putc(BEL);
-		for (i=1; (p = g.result[i]); i++) {
+		/* No annoying bell here... (at&t ksh doesn't do this either)
+		if (g.result[1] != 0)
+			vi_error();
+		*/
+		for (i = 1; (p = g.result[i]); i++) {
 			int j;
 
 			for (j = 0; j < len; j++)
@@ -1957,7 +2102,7 @@ complete_word(command)
 					break;
 				}
 		}
-		if (putbuf(g.result[0], len, 0) != 0) 
+		if (putbuf(g.result[0], len, 0) != 0)
 			rval = -1;
 	}
 	/* Check to see if it's a directory (len == 0 probably isn't needed) */
@@ -1991,10 +2136,15 @@ print_expansions(command)
 		x_putc(' ');
 	while (*p && **p) {
 		char 	*q;
-		if ((q = strrchr_dirsep(*p)) == 0)
-			x_puts(*p);
-		else
-			x_puts(++q);
+
+		/* Skip trailing /'s (set -o markdirs creates them) */
+		for (q = *p + strlen(*p) - 1; q > *p && ISDIRSEP(*q); --q)
+			;
+		for (; q > *p && !ISDIRSEP(*q); --q)
+			;
+		if (ISDIRSEP(*q) && q[1])
+			q++;
+		x_puts(q);
 		x_putc(' ');
 		p++;
 	}
@@ -2036,4 +2186,22 @@ x_vi_zotc(c)
 	x_putc(c);
 }
 
+static void
+vi_error()
+{
+	/* Beem out of any macros as soon as an error occurs */
+	vi_macro_reset();
+	x_putc(BEL);
+	x_flush();
+}
+
+static void
+vi_macro_reset()
+{
+	if (macro.p) {
+		afree(macro.buf, APERM);
+		memset((char *) &macro, 0, sizeof(macro));
+	}
+}
+
 #endif	/* VI */
